# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz pull."""

import os
import sys

from breezy import branch, debug, osutils, tests, uncommit, urlutils, workingtree
from breezy.bzr import remote
from breezy.directory_service import directories
from breezy.tests import fixtures, script


class TestPull(tests.TestCaseWithTransport):
    def example_branch(self, path="."):
        tree = self.make_branch_and_tree(path)
        self.build_tree_contents(
            [
                (osutils.pathjoin(path, "hello"), b"foo"),
                (osutils.pathjoin(path, "goodbye"), b"baz"),
            ]
        )
        tree.add("hello")
        tree.commit(message="setup")
        tree.add("goodbye")
        tree.commit(message="setup")
        return tree

    def test_pull(self):
        """Pull changes from one branch to another."""
        a_tree = self.example_branch("a")
        base_rev = a_tree.branch.last_revision()
        self.run_bzr("pull", retcode=3, working_dir="a")
        self.run_bzr("missing", retcode=3, working_dir="a")
        self.run_bzr("missing .", working_dir="a")
        self.run_bzr("missing", working_dir="a")
        # this will work on windows because we check for the same branch
        # in pull - if it fails, it is a regression
        self.run_bzr("pull", working_dir="a")
        self.run_bzr("pull /", retcode=3, working_dir="a")
        if sys.platform not in ("win32", "cygwin"):
            self.run_bzr("pull", working_dir="a")

        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.run_bzr("pull", working_dir="b")
        os.mkdir("b/subdir")
        b_tree.add("subdir")
        new_rev = b_tree.commit(message="blah", allow_pointless=True)

        a = branch.Branch.open("a")
        b = branch.Branch.open("b")
        self.assertEqual(a.last_revision(), base_rev)
        self.assertEqual(b.last_revision(), new_rev)

        self.run_bzr("pull ../b", working_dir="a")
        self.assertEqual(a.last_revision(), b.last_revision())
        a_tree.commit(message="blah2", allow_pointless=True)
        b_tree.commit(message="blah3", allow_pointless=True)
        # no overwrite
        self.run_bzr("pull ../a", retcode=3, working_dir="b")
        b_tree.controldir.sprout("overwriteme")
        self.run_bzr("pull --overwrite ../a", working_dir="overwriteme")
        overwritten = branch.Branch.open("overwriteme")
        self.assertEqual(overwritten.last_revision(), a.last_revision())
        a_tree.merge_from_branch(b_tree.branch)
        a_tree.commit(message="blah4", allow_pointless=True)

        self.run_bzr("pull ../../a", working_dir="b/subdir")
        self.assertEqual(a.last_revision(), b.last_revision())
        sub_tree = workingtree.WorkingTree.open_containing("b/subdir")[0]
        sub_tree.commit(message="blah5", allow_pointless=True)
        sub_tree.commit(message="blah6", allow_pointless=True)
        self.run_bzr("pull ../a", working_dir="b")
        a_tree.commit(message="blah7", allow_pointless=True)
        a_tree.merge_from_branch(b_tree.branch)
        a_tree.commit(message="blah8", allow_pointless=True)
        self.run_bzr("pull ../b", working_dir="a")
        self.run_bzr("pull ../b", working_dir="a")

    def test_pull_dash_d(self):
        self.example_branch("a")
        self.make_branch_and_tree("b")
        self.make_branch_and_tree("c")
        # pull into that branch
        self.run_bzr("pull -d b a")
        # pull into a branch specified by a url
        c_url = urlutils.local_path_to_url("c")
        self.assertStartsWith(c_url, "file://")
        self.run_bzr(["pull", "-d", c_url, "a"])

    def test_pull_revision(self):
        """Pull some changes from one branch to another."""
        a_tree = self.example_branch("a")
        self.build_tree_contents([("a/hello2", b"foo"), ("a/goodbye2", b"baz")])
        a_tree.add("hello2")
        a_tree.commit(message="setup")
        a_tree.add("goodbye2")
        a_tree.commit(message="setup")

        a_tree.controldir.sprout(
            "b", revision_id=a_tree.branch.get_rev_id(1)
        ).open_workingtree()
        self.run_bzr("pull -r 2", working_dir="b")
        a = branch.Branch.open("a")
        b = branch.Branch.open("b")
        self.assertEqual(a.revno(), 4)
        self.assertEqual(b.revno(), 2)
        self.run_bzr("pull -r 3", working_dir="b")
        self.assertEqual(b.revno(), 3)
        self.run_bzr("pull -r 4", working_dir="b")
        self.assertEqual(a.last_revision(), b.last_revision())

    def test_pull_tags(self):
        """Tags are updated by pull, and revisions named in those tags are
        fetched.
        """
        # Make a source, sprout a target off it
        builder = self.make_branch_builder("source")
        source, rev1, rev2 = fixtures.build_branch_with_non_ancestral_rev(builder)
        source.get_config_stack().set("branch.fetch_tags", True)
        target_bzrdir = source.controldir.sprout("target")
        source.tags.set_tag("tag-a", rev2)
        # Pull from source
        self.run_bzr("pull -d target source")
        target = target_bzrdir.open_branch()
        # The tag is present, and so is its revision.
        self.assertEqual(rev2, target.tags.lookup_tag("tag-a"))
        target.repository.get_revision(rev2)

    def test_overwrite_uptodate(self):
        # Make sure pull --overwrite overwrites
        # even if the target branch has merged
        # everything already.
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/foo", b"original\n")])
        a_tree.add("foo")
        a_tree.commit(message="initial commit")

        b_tree = a_tree.controldir.sprout("b").open_workingtree()

        self.build_tree_contents([("a/foo", b"changed\n")])
        a_tree.commit(message="later change")

        self.build_tree_contents([("a/foo", b"a third change")])
        a_tree.commit(message="a third change")

        self.assertEqual(a_tree.branch.last_revision_info()[0], 3)

        b_tree.merge_from_branch(a_tree.branch)
        b_tree.commit(message="merge")

        self.assertEqual(b_tree.branch.last_revision_info()[0], 2)

        self.run_bzr("pull --overwrite ../a", working_dir="b")
        (last_revinfo_b) = b_tree.branch.last_revision_info()
        self.assertEqual(last_revinfo_b[0], 3)
        self.assertEqual(last_revinfo_b[1], a_tree.branch.last_revision())

    def test_overwrite_children(self):
        # Make sure pull --overwrite sets the revision-history
        # to be identical to the pull source, even if we have convergence
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/foo", b"original\n")])
        a_tree.add("foo")
        a_tree.commit(message="initial commit")

        b_tree = a_tree.controldir.sprout("b").open_workingtree()

        self.build_tree_contents([("a/foo", b"changed\n")])
        a_tree.commit(message="later change")

        self.build_tree_contents([("a/foo", b"a third change")])
        a_tree.commit(message="a third change")

        self.assertEqual(a_tree.branch.last_revision_info()[0], 3)

        b_tree.merge_from_branch(a_tree.branch)
        b_tree.commit(message="merge")

        self.assertEqual(b_tree.branch.last_revision_info()[0], 2)

        self.build_tree_contents([("a/foo", b"a fourth change\n")])
        a_tree.commit(message="a fourth change")

        rev_info_a = a_tree.branch.last_revision_info()
        self.assertEqual(rev_info_a[0], 4)

        # With convergence, we could just pull over the
        # new change, but with --overwrite, we want to switch our history
        self.run_bzr("pull --overwrite ../a", working_dir="b")
        rev_info_b = b_tree.branch.last_revision_info()
        self.assertEqual(rev_info_b[0], 4)
        self.assertEqual(rev_info_b, rev_info_a)

    def test_pull_remember(self):
        """Pull changes from one branch to another and test parent location."""
        t = self.get_transport()
        tree_a = self.make_branch_and_tree("branch_a")
        branch_a = tree_a.branch
        self.build_tree(["branch_a/a"])
        tree_a.add("a")
        tree_a.commit("commit a")
        tree_b = branch_a.controldir.sprout("branch_b").open_workingtree()
        branch_b = tree_b.branch
        tree_c = branch_a.controldir.sprout("branch_c").open_workingtree()
        branch_c = tree_c.branch
        self.build_tree(["branch_a/b"])
        tree_a.add("b")
        tree_a.commit("commit b")
        # reset parent
        parent = branch_b.get_parent()
        branch_b = branch.Branch.open("branch_b")
        branch_b.set_parent(None)
        self.assertEqual(None, branch_b.get_parent())
        # test pull for failure without parent set
        out = self.run_bzr("pull", retcode=3, working_dir="branch_b")
        self.assertEqual(
            out, ("", "brz: ERROR: No pull location known or specified.\n")
        )
        # test implicit --remember when no parent set, this pull conflicts
        self.build_tree(["branch_b/d"])
        tree_b.add("d")
        tree_b.commit("commit d")
        out = self.run_bzr("pull ../branch_a", retcode=3, working_dir="branch_b")
        self.assertEqual(
            out,
            (
                "",
                "brz: ERROR: These branches have diverged."
                " Use the missing command to see how.\n"
                "Use the merge command to reconcile them.\n",
            ),
        )
        tree_b = tree_b.controldir.open_workingtree()
        branch_b = tree_b.branch
        self.assertEqual(parent, branch_b.get_parent())
        # test implicit --remember after resolving previous failure
        uncommit.uncommit(branch=branch_b, tree=tree_b)
        t.delete("branch_b/d")
        self.run_bzr("pull", working_dir="branch_b")
        # Refresh the branch object as 'pull' modified it
        branch_b = branch_b.controldir.open_branch()
        self.assertEqual(branch_b.get_parent(), parent)
        # test explicit --remember
        self.run_bzr("pull ../branch_c --remember", working_dir="branch_b")
        # Refresh the branch object as 'pull' modified it
        branch_b = branch_b.controldir.open_branch()
        self.assertEqual(branch_c.controldir.root_transport.base, branch_b.get_parent())

    def test_pull_bundle(self):
        from breezy.bzr.testament import Testament

        # Build up 2 trees and prepare for a pull
        tree_a = self.make_branch_and_tree("branch_a")
        with open("branch_a/a", "wb") as f:
            f.write(b"hello")
        tree_a.add("a")
        tree_a.commit("message")

        tree_b = tree_a.controldir.sprout("branch_b").open_workingtree()

        # Make a change to 'a' that 'b' can pull
        with open("branch_a/a", "wb") as f:
            f.write(b"hey there")
        tree_a.commit("message")

        # Create the bundle for 'b' to pull
        self.run_bzr("bundle ../branch_b -o ../bundle", working_dir="branch_a")

        out, err = self.run_bzr("pull ../bundle", working_dir="branch_b")
        self.assertEqual(out, "Now on revision 2.\n")
        self.assertEqual(err, " M  a\nAll changes applied successfully.\n")

        self.assertEqualDiff(
            tree_a.branch.last_revision(), tree_b.branch.last_revision()
        )

        testament_a = Testament.from_revision(
            tree_a.branch.repository, tree_a.get_parent_ids()[0]
        )
        testament_b = Testament.from_revision(
            tree_b.branch.repository, tree_b.get_parent_ids()[0]
        )
        self.assertEqualDiff(testament_a.as_text(), testament_b.as_text())

        # it is legal to attempt to pull an already-merged bundle
        out, err = self.run_bzr("pull ../bundle", working_dir="branch_b")
        self.assertEqual(err, "")
        self.assertEqual(out, "No revisions or tags to pull.\n")

    def test_pull_verbose_no_files(self):
        """Pull --verbose should not list modified files."""
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree(["tree_a/foo"])
        tree_a.add("foo")
        tree_a.commit("bar")
        self.make_branch_and_tree("tree_b")
        out = self.run_bzr("pull --verbose -d tree_b tree_a")[0]
        self.assertContainsRe(out, "bar")
        self.assertNotContainsRe(out, "added:")
        self.assertNotContainsRe(out, "foo")

    def test_pull_quiet(self):
        """Check that brz pull --quiet does not print anything."""
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree(["tree_a/foo"])
        tree_a.add("foo")
        revision_id = tree_a.commit("bar")
        tree_b = tree_a.controldir.sprout("tree_b").open_workingtree()
        out, err = self.run_bzr("pull --quiet -d tree_b")
        self.assertEqual(out, "")
        self.assertEqual(err, "")
        self.assertEqual(tree_b.last_revision(), revision_id)
        self.build_tree(["tree_a/moo"])
        tree_a.add("moo")
        revision_id = tree_a.commit("quack")
        out, err = self.run_bzr("pull --quiet -d tree_b")
        self.assertEqual(out, "")
        self.assertEqual(err, "")
        self.assertEqual(tree_b.last_revision(), revision_id)

    def test_pull_from_directory_service(self):
        source = self.make_branch_and_tree("source")
        source.commit("commit 1")
        target = source.controldir.sprout("target").open_workingtree()
        source_last = source.commit("commit 2")

        class FooService:
            """A directory service that always returns source."""

            def look_up(self, name, url, purpose=None):
                return "source"

        directories.register("foo:", FooService, "Testing directory service")
        self.addCleanup(directories.remove, "foo:")
        self.run_bzr("pull foo:bar -d target")
        self.assertEqual(source_last, target.last_revision())

    def test_pull_verbose_defaults_to_long(self):
        self.example_branch("source")
        self.make_branch_and_tree("target")
        out = self.run_bzr("pull -v source -d target")[0]
        self.assertContainsRe(out, r"revno: 1\ncommitter: .*\nbranch nick: source")
        self.assertNotContainsRe(out, r"\n {4}1 .*\n {6}setup\n")

    def test_pull_verbose_uses_default_log(self):
        self.example_branch("source")
        target = self.make_branch_and_tree("target")
        target.branch.get_config_stack().set("log_format", "short")
        out = self.run_bzr("pull -v source -d target")[0]
        self.assertContainsRe(out, r"\n {4}1 .*\n {6}setup\n")
        self.assertNotContainsRe(out, r"revno: 1\ncommitter: .*\nbranch nick: source")

    def test_pull_smart_bound_branch(self):
        self.setup_smart_server_with_call_log()
        parent = self.make_branch_and_tree("parent")
        parent.commit(message="first commit")
        child = parent.controldir.sprout("child").open_workingtree()
        child.commit(message="second commit")
        parent.branch.create_checkout("checkout")
        self.run_bzr(["pull", self.get_url("child")], working_dir="checkout")

    def test_pull_smart_stacked_streaming_acceptance(self):
        """'brz pull -r 123' works on stacked, smart branches, even when the
        revision specified by the revno is only present in the fallback
        repository.

        See <https://launchpad.net/bugs/380314>
        """
        self.setup_smart_server_with_call_log()
        # Make a stacked-on branch with two commits so that the
        # revision-history can't be determined just by looking at the parent
        # field in the revision in the stacked repo.
        parent = self.make_branch_and_tree("parent", format="1.9")
        parent.commit(message="first commit")
        parent.commit(message="second commit")
        local = parent.controldir.sprout("local").open_workingtree()
        local.commit(message="local commit")
        local.branch.create_clone_on_transport(
            self.get_transport("stacked"), stacked_on=self.get_url("parent")
        )
        self.make_branch_and_tree("empty", format="1.9")
        self.reset_smart_call_log()
        self.run_bzr(["pull", "-r", "1", self.get_url("stacked")], working_dir="empty")
        # This figure represent the amount of work to perform this use case. It
        # is entirely ok to reduce this number if a test fails due to rpc_count
        # being too low. If rpc_count increases, more network roundtrips have
        # become necessary for this use case. Please do not adjust this number
        # upwards without agreement from bzr's network support maintainers.
        self.assertLength(20, self.hpss_calls)
        self.assertLength(1, self.hpss_connections)
        remote = branch.Branch.open("stacked")
        self.assertEndsWith(remote.get_stacked_on_url(), "/parent")

    def test_pull_cross_format_warning(self):
        """You get a warning for probably slow cross-format pulls."""
        # this is assumed to be going through InterDifferingSerializer
        from_tree = self.make_branch_and_tree("from", format="2a")
        self.make_branch_and_tree("to", format="1.14-rich-root")
        from_tree.commit(message="first commit")
        out, err = self.run_bzr(["pull", "-d", "to", "from"])
        self.assertContainsRe(err, "(?m)Doing on-the-fly conversion")

    def test_pull_cross_format_warning_no_IDS(self):
        """You get a warning for probably slow cross-format pulls."""
        # this simulates what would happen across the network, where
        # interdifferingserializer is not active

        debug.debug_flags.add("IDS_never")
        # TestCase take care of restoring them

        from_tree = self.make_branch_and_tree("from", format="2a")
        self.make_branch_and_tree("to", format="1.14-rich-root")
        from_tree.commit(message="first commit")
        out, err = self.run_bzr(["pull", "-d", "to", "from"])
        self.assertContainsRe(err, "(?m)Doing on-the-fly conversion")

    def test_pull_cross_format_from_network(self):
        self.setup_smart_server_with_call_log()
        from_tree = self.make_branch_and_tree("from", format="2a")
        self.make_branch_and_tree("to", format="1.14-rich-root")
        self.assertIsInstance(from_tree.branch, remote.RemoteBranch)
        from_tree.commit(message="first commit")
        out, err = self.run_bzr(
            ["pull", "-d", "to", from_tree.branch.controldir.root_transport.base]
        )
        self.assertContainsRe(err, "(?m)Doing on-the-fly conversion")

    def test_pull_to_experimental_format_warning(self):
        """You get a warning for pulling into experimental formats."""
        from_tree = self.make_branch_and_tree("from", format="development-subtree")
        self.make_branch_and_tree("to", format="development-subtree")
        from_tree.commit(message="first commit")
        out, err = self.run_bzr(["pull", "-d", "to", "from"])
        self.assertContainsRe(err, "(?m)Fetching into experimental format")

    def test_pull_cross_to_experimental_format_warning(self):
        """You get a warning for pulling into experimental formats."""
        from_tree = self.make_branch_and_tree("from", format="2a")
        self.make_branch_and_tree("to", format="development-subtree")
        from_tree.commit(message="first commit")
        out, err = self.run_bzr(["pull", "-d", "to", "from"])
        self.assertContainsRe(err, "(?m)Fetching into experimental format")

    def test_pull_show_base(self):
        """Brz pull supports --show-base.

        see https://bugs.launchpad.net/bzr/+bug/202374
        """
        # create two trees with conflicts, setup conflict, check that
        # conflicted file looks correct
        a_tree = self.example_branch("a")
        a_tree.controldir.sprout("b").open_workingtree()

        with open(osutils.pathjoin("a", "hello"), "w") as f:
            f.write("fee")
        a_tree.commit("fee")

        with open(osutils.pathjoin("b", "hello"), "w") as f:
            f.write("fie")

        out, err = self.run_bzr(["pull", "-d", "b", "a", "--show-base"])

        # check for message here
        self.assertEqual(
            err, " M  hello\nText conflict in hello\n1 conflicts encountered.\n"
        )

        with open(osutils.pathjoin("b", "hello")) as f:
            self.assertEqualDiff(
                "<<<<<<< TREE\n"
                "fie||||||| BASE-REVISION\n"
                "foo=======\n"
                "fee>>>>>>> MERGE-SOURCE\n",
                f.read(),
            )

    def test_pull_warns_about_show_base_when_no_working_tree(self):
        """--show-base is useless if there's no working tree.

        see https://bugs.launchpad.net/bzr/+bug/1022160
        """
        self.make_branch("from")
        self.make_branch("to")
        out = self.run_bzr(["pull", "-d", "to", "from", "--show-base"])
        self.assertEqual(
            out,
            (
                "No revisions or tags to pull.\n",
                "No working tree, ignoring --show-base\n",
            ),
        )

    def test_pull_tag_conflicts(self):
        """Pulling tags with conflicts will change the exit code."""
        # create a branch, see that --show-base fails
        from_tree = self.make_branch_and_tree("from")
        from_tree.branch.tags.set_tag("mytag", b"somerevid")
        to_tree = self.make_branch_and_tree("to")
        to_tree.branch.tags.set_tag("mytag", b"anotherrevid")
        out = self.run_bzr(["pull", "-d", "to", "from"], retcode=1)
        self.assertEqual(
            out, ("No revisions to pull.\nConflicting tags:\n    mytag\n", "")
        )

    def test_pull_tag_notification(self):
        """Pulling tags with conflicts will change the exit code."""
        # create a branch, see that --show-base fails
        from_tree = self.make_branch_and_tree("from")
        from_tree.branch.tags.set_tag("mytag", b"somerevid")
        self.make_branch_and_tree("to")
        out = self.run_bzr(["pull", "-d", "to", "from"])
        self.assertEqual(out, ("1 tag(s) updated.\n", ""))

    def test_overwrite_tags(self):
        """--overwrite-tags only overwrites tags, not revisions."""
        from_tree = self.make_branch_and_tree("from")
        from_tree.branch.tags.set_tag("mytag", b"somerevid")
        to_tree = self.make_branch_and_tree("to")
        to_tree.branch.tags.set_tag("mytag", b"anotherrevid")
        revid1 = to_tree.commit("my commit")
        out = self.run_bzr(["pull", "-d", "to", "from"], retcode=1)
        self.assertEqual(
            out, ("No revisions to pull.\nConflicting tags:\n    mytag\n", "")
        )
        out = self.run_bzr(["pull", "-d", "to", "--overwrite-tags", "from"])
        self.assertEqual(out, ("1 tag(s) updated.\n", ""))

        self.assertEqual(to_tree.branch.tags.lookup_tag("mytag"), b"somerevid")
        self.assertEqual(to_tree.branch.last_revision(), revid1)

    def test_pull_tag_overwrite(self):
        """Pulling tags with --overwrite only reports changed tags."""
        # create a branch, see that --show-base fails
        from_tree = self.make_branch_and_tree("from")
        from_tree.branch.tags.set_tag("mytag", b"somerevid")
        to_tree = self.make_branch_and_tree("to")
        to_tree.branch.tags.set_tag("mytag", b"somerevid")
        out = self.run_bzr(["pull", "--overwrite", "-d", "to", "from"])
        self.assertEqual(out, ("No revisions or tags to pull.\n", ""))


class TestPullOutput(script.TestCaseWithTransportAndScript):
    def test_pull_log_format(self):
        self.run_script("""
            $ brz init trunk
            Created a standalone tree (format: 2a)
            $ cd trunk
            $ echo foo > file
            $ brz add
            adding file
            $ brz commit -m 'we need some foo'
            2>Committing to:...trunk/
            2>added file
            2>Committed revision 1.
            $ cd ..
            $ brz init feature
            Created a standalone tree (format: 2a)
            $ cd feature
            $ brz pull -v ../trunk -Olog_format=line
            Now on revision 1.
            Added Revisions:
            1: jrandom@example.com ...we need some foo
            2>+N  file
            2>All changes applied successfully.
            """)



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import branch
from breezy.bzr import vf_search
from breezy.tests.per_repository import TestCaseWithRepository


class TestFetchBase(TestCaseWithRepository):
    def make_source_branch(self):
        # It would be nice if there was a way to force this to be memory-only
        builder = self.make_branch_builder("source")
        content = [
            b"content lines\n"
            b"for the first revision\n"
            b"which is a marginal amount of content\n"
        ]
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("a", b"a-id", "file", b"".join(content))),
            ],
            revision_id=b"A-id",
        )
        content.append(b"and some more lines for B\n")
        builder.build_snapshot(
            [b"A-id"], [("modify", ("a", b"".join(content)))], revision_id=b"B-id"
        )
        content.append(b"and yet even more content for C\n")
        builder.build_snapshot(
            [b"B-id"], [("modify", ("a", b"".join(content)))], revision_id=b"C-id"
        )
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        return content, source_b


class TestFetch(TestFetchBase):
    def test_sprout_from_stacked_with_short_history(self):
        content, source_b = self.make_source_branch()
        # Split the generated content into a base branch, and a stacked branch
        # Use 'make_branch' which gives us a bzr:// branch when appropriate,
        # rather than creating a branch-on-disk
        stack_b = self.make_branch("stack-on")
        stack_b.pull(source_b, stop_revision=b"B-id")
        target_b = self.make_branch("target")
        target_b.set_stacked_on_url("../stack-on")
        target_b.pull(source_b, stop_revision=b"C-id")
        # At this point, we should have a target branch, with 1 revision, on
        # top of the source.
        final_b = self.make_branch("final")
        final_b.pull(target_b)
        final_b.lock_read()
        self.addCleanup(final_b.unlock)
        self.assertEqual(b"C-id", final_b.last_revision())
        text_keys = [(b"a-id", b"A-id"), (b"a-id", b"B-id"), (b"a-id", b"C-id")]
        stream = final_b.repository.texts.get_record_stream(
            text_keys, "unordered", True
        )
        records = sorted([(r.key, r.get_bytes_as("fulltext")) for r in stream])
        self.assertEqual(
            [
                ((b"a-id", b"A-id"), b"".join(content[:-2])),
                ((b"a-id", b"B-id"), b"".join(content[:-1])),
                ((b"a-id", b"C-id"), b"".join(content)),
            ],
            records,
        )

    def test_sprout_from_smart_stacked_with_short_history(self):
        content, source_b = self.make_source_branch()
        transport = self.make_smart_server("server")
        transport.ensure_base()
        url = transport.abspath("")
        source_b.controldir.sprout(url + "/stack-on", revision_id=b"B-id")
        # self.make_branch only takes relative paths, so we do it the 'hard'
        # way
        target_transport = transport.clone("target")
        target_transport.ensure_base()
        target_bzrdir = self.bzrdir_format.initialize_on_transport(target_transport)
        target_bzrdir.create_repository()
        target_b = target_bzrdir.create_branch()
        target_b.set_stacked_on_url("../stack-on")
        target_b.pull(source_b, stop_revision=b"C-id")
        # Now we should be able to branch from the remote location to a local
        # location
        final_b = target_b.controldir.sprout("final").open_branch()
        self.assertEqual(b"C-id", final_b.last_revision())

        # bzrdir.sprout() has slightly different code paths if you supply a
        # revision_id versus not. If you supply revision_id, then you get a
        # PendingAncestryResult for the search, versus a SearchResult...
        target_b.controldir.sprout("final2", revision_id=b"C-id").open_branch()
        self.assertEqual(b"C-id", final_b.last_revision())

    def make_source_with_ghost_and_stacked_target(self):
        builder = self.make_branch_builder("source")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("file", b"file-id", "file", b"content\n")),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot([b"A-id", b"ghost-id"], [], revision_id=b"B-id")
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        base = self.make_branch("base")
        base.pull(source_b, stop_revision=b"A-id")
        stacked = self.make_branch("stacked")
        stacked.set_stacked_on_url("../base")
        return source_b, base, stacked

    def test_fetch_with_ghost_stacked(self):
        (source_b, base, stacked) = self.make_source_with_ghost_and_stacked_target()
        stacked.pull(source_b, stop_revision=b"B-id")

    def test_fetch_into_smart_stacked_with_ghost(self):
        (source_b, base, stacked) = self.make_source_with_ghost_and_stacked_target()
        # Now, create a smart server on 'stacked' and re-open to force the
        # target to be a smart target
        trans = self.make_smart_server("stacked")
        stacked = branch.Branch.open(trans.base)
        stacked.lock_write()
        self.addCleanup(stacked.unlock)
        stacked.pull(source_b, stop_revision=b"B-id")

    def test_fetch_to_stacked_from_smart_with_ghost(self):
        (source_b, base, stacked) = self.make_source_with_ghost_and_stacked_target()
        # Now, create a smart server on 'source' and re-open to force the
        # target to be a smart target
        trans = self.make_smart_server("source")
        source_b = branch.Branch.open(trans.base)
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        stacked.pull(source_b, stop_revision=b"B-id")


class TestFetchFromRepoWithUnconfiguredFallbacks(TestFetchBase):
    def make_stacked_source_repo(self):
        _, source_b = self.make_source_branch()
        # Use 'make_branch' which gives us a bzr:// branch when appropriate,
        # rather than creating a branch-on-disk
        stack_b = self.make_branch("stack-on")
        stack_b.pull(source_b, stop_revision=b"B-id")
        stacked_b = self.make_branch("stacked")
        stacked_b.set_stacked_on_url("../stack-on")
        stacked_b.pull(source_b, stop_revision=b"C-id")
        return stacked_b.repository

    def test_fetch_everything_includes_parent_invs(self):
        stacked = self.make_stacked_source_repo()
        repo_missing_fallbacks = stacked.controldir.open_repository()
        self.addCleanup(repo_missing_fallbacks.lock_read().unlock)
        target = self.make_repository("target")
        self.addCleanup(target.lock_write().unlock)
        target.fetch(
            repo_missing_fallbacks,
            fetch_spec=vf_search.EverythingResult(repo_missing_fallbacks),
        )
        self.assertEqual(
            repo_missing_fallbacks.revisions.keys(), target.revisions.keys()
        )
        self.assertEqual(
            repo_missing_fallbacks.inventories.keys(), target.inventories.keys()
        )
        self.assertEqual([b"C-id"], sorted(k[-1] for k in target.revisions.keys()))
        self.assertEqual(
            [b"B-id", b"C-id"], sorted(k[-1] for k in target.inventories.keys())
        )



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of inventories of working trees."""

import os

from breezy import tests
from breezy.bzr import inventory
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestRevert(TestCaseWithWorkingTree):
    def test_dangling_id(self):
        wt = self.make_branch_and_tree("b1")
        wt.lock_tree_write()
        self.addCleanup(wt.unlock)
        self.assertEqual(len(list(wt.all_versioned_paths())), 1)
        with open("b1/a", "wb") as f:
            f.write(b"a test\n")
        wt.add("a")
        self.assertEqual(len(list(wt.all_versioned_paths())), 2)
        wt.flush()  # workaround revert doing wt._write_inventory for now.
        os.unlink("b1/a")
        wt.revert()
        self.assertEqual(len(list(wt.all_versioned_paths())), 1)


class TestApplyInventoryDelta(TestCaseWithWorkingTree):
    def setUp(self):
        super().setUp()
        if not self.bzrdir_format.repository_format.supports_full_versioned_files:
            raise tests.TestNotApplicable("format does not support inventory deltas")

    def test_add(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.addCleanup(wt.unlock)
        root_id = wt.path2id("")
        wt.apply_inventory_delta(
            [
                (
                    None,
                    "bar/foo",
                    b"foo-id",
                    inventory.InventoryFile(b"foo-id", "foo", parent_id=b"bar-id"),
                ),
                (
                    None,
                    "bar",
                    b"bar-id",
                    inventory.InventoryDirectory(b"bar-id", "bar", parent_id=root_id),
                ),
            ]
        )
        self.assertEqual("bar/foo", wt.id2path(b"foo-id"))
        self.assertEqual("bar", wt.id2path(b"bar-id"))

    def test_remove(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(["foo/", "foo/bar"])
        wt.add(["foo", "foo/bar"], ids=[b"foo-id", b"bar-id"])
        wt.apply_inventory_delta(
            [("foo", None, b"foo-id", None), ("foo/bar", None, b"bar-id", None)]
        )
        self.assertFalse(wt.is_versioned("foo"))

    def test_rename_dir_with_children(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        root_id = wt.path2id("")
        self.addCleanup(wt.unlock)
        self.build_tree(["foo/", "foo/bar"])
        wt.add(["foo", "foo/bar"], ids=[b"foo-id", b"bar-id"])
        wt.apply_inventory_delta(
            [
                (
                    "foo",
                    "baz",
                    b"foo-id",
                    inventory.InventoryDirectory(b"foo-id", "baz", root_id),
                )
            ]
        )
        # foo/bar should have been followed the rename of its parent to baz/bar
        self.assertEqual("baz", wt.id2path(b"foo-id"))
        self.assertEqual("baz/bar", wt.id2path(b"bar-id"))

    def test_rename_dir_with_children_with_children(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        root_id = wt.path2id("")
        self.addCleanup(wt.unlock)
        self.build_tree(["foo/", "foo/bar/", "foo/bar/baz"])
        wt.add(["foo", "foo/bar", "foo/bar/baz"], ids=[b"foo-id", b"bar-id", b"baz-id"])
        wt.apply_inventory_delta(
            [
                (
                    "foo",
                    "quux",
                    b"foo-id",
                    inventory.InventoryDirectory(b"foo-id", "quux", root_id),
                )
            ]
        )
        # foo/bar/baz should have been followed the rename of its parent's
        # parent to quux/bar/baz
        self.assertEqual("quux/bar/baz", wt.id2path(b"baz-id"))

    def test_rename_file(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.addCleanup(wt.unlock)
        self.build_tree(["foo/", "foo/bar", "baz/"])
        wt.add(["foo", "foo/bar", "baz"], ids=[b"foo-id", b"bar-id", b"baz-id"])
        wt.apply_inventory_delta(
            [
                (
                    "foo/bar",
                    "baz/bar",
                    b"bar-id",
                    inventory.InventoryFile(b"bar-id", "bar", b"baz-id"),
                )
            ]
        )
        self.assertEqual("baz/bar", wt.id2path(b"bar-id"))

    def test_rename_swap(self):
        """Test the swap-names edge case.

        foo and bar should swap names, but retain their children.  If this
        works, any simpler rename ought to work.
        """
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        root_id = wt.path2id("")
        self.addCleanup(wt.unlock)
        self.build_tree(["foo/", "foo/bar", "baz/", "baz/qux"])
        wt.add(
            ["foo", "foo/bar", "baz", "baz/qux"],
            ids=[b"foo-id", b"bar-id", b"baz-id", b"qux-id"],
        )
        wt.apply_inventory_delta(
            [
                (
                    "foo",
                    "baz",
                    b"foo-id",
                    inventory.InventoryDirectory(b"foo-id", "baz", root_id),
                ),
                (
                    "baz",
                    "foo",
                    b"baz-id",
                    inventory.InventoryDirectory(b"baz-id", "foo", root_id),
                ),
            ]
        )
        self.assertEqual("baz/bar", wt.id2path(b"bar-id"))
        self.assertEqual("foo/qux", wt.id2path(b"qux-id"))

    def test_child_rename_ordering(self):
        """Test the rename-parent, move child edge case.

        (A naive implementation may move the parent first, and then be
         unable to find the child.)
        """
        wt = self.make_branch_and_tree(".")
        root_id = wt.path2id("")
        self.build_tree(["dir/", "dir/child", "other/"])
        wt.add(["dir", "dir/child", "other"], ids=[b"dir-id", b"child-id", b"other-id"])
        # this delta moves dir-id to dir2 and reparents
        # child-id to a parent of other-id
        wt.apply_inventory_delta(
            [
                (
                    "dir",
                    "dir2",
                    b"dir-id",
                    inventory.InventoryDirectory(b"dir-id", "dir2", root_id),
                ),
                (
                    "dir/child",
                    "other/child",
                    b"child-id",
                    inventory.InventoryFile(b"child-id", "child", b"other-id"),
                ),
            ]
        )
        self.assertEqual("dir2", wt.id2path(b"dir-id"))
        self.assertEqual("other/child", wt.id2path(b"child-id"))

    def test_replace_root(self):
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.addCleanup(wt.unlock)

        root_id = wt.path2id("")
        wt.apply_inventory_delta(
            [
                ("", None, root_id, None),
                (
                    None,
                    "",
                    b"root-id",
                    inventory.InventoryDirectory(b"root-id", "", None),
                ),
            ]
        )


class TestTreeReference(TestCaseWithWorkingTree):
    def test_tree_reference_matches_inv(self):
        base = self.make_branch_and_tree("base")
        if base.branch.repository._format.supports_full_versioned_files:
            raise tests.TestNotApplicable("format does not support inventory deltas")
        if not base.supports_tree_reference():
            raise tests.TestNotApplicable("wt doesn't support nested trees")
        if base.has_versioned_directories():
            # We add it as a directory, but it becomes a tree-reference
            base.add(["subdir"], ["directory"])
            subdir = self.make_branch_and_tree("base/subdir")
        else:
            subdir = self.make_branch_and_tree("base/subdir")
            subdir.commit("")
            # We add it as a directory, but it becomes a tree-reference
            base.add(["subdir"], ["tree-reference"])
        self.addCleanup(base.lock_read().unlock)
        # Note: we aren't strict about ie.kind being 'directory' here, what we
        # are strict about is that wt.inventory should match
        # wt.current_dirstate()'s idea about what files are where.
        path, ie = next(base.iter_entries_by_dir(specific_files=["subdir"]))
        self.assertEqual("subdir", path)
        self.assertEqual("tree-reference", ie.kind)



# Copyright (C) 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the python and pyrex extensions of KnownGraph."""

import pprint

from .. import _known_graph_py, errors, tests
from ..revision import NULL_REVISION
from . import features, test_graph
from .scenarios import load_tests_apply_scenarios


def caching_scenarios():
    scenarios = [
        ("python", {"module": _known_graph_py, "do_cache": True}),
    ]
    if compiled_known_graph_feature.available():
        scenarios.append(
            ("C", {"module": compiled_known_graph_feature.module, "do_cache": True})
        )
    return scenarios


def non_caching_scenarios():
    scenarios = [
        ("python-nocache", {"module": _known_graph_py, "do_cache": False}),
    ]
    if compiled_known_graph_feature.available():
        scenarios.append(
            (
                "C-nocache",
                {"module": compiled_known_graph_feature.module, "do_cache": False},
            )
        )
    return scenarios


load_tests = load_tests_apply_scenarios


compiled_known_graph_feature = features.ModuleAvailableFeature(
    "breezy._known_graph_pyx"
)


#  a
#  |\
#  b |
#  | |
#  c |
#   \|
#    d
alt_merge = {b"a": [], b"b": [b"a"], b"c": [b"b"], b"d": [b"a", b"c"]}


class TestCaseWithKnownGraph(tests.TestCase):
    scenarios = caching_scenarios()
    module = None  # Set by load_tests

    def make_known_graph(self, ancestry):
        return self.module.KnownGraph(ancestry, do_cache=self.do_cache)


class TestKnownGraph(TestCaseWithKnownGraph):
    def assertGDFO(self, graph, rev, gdfo):
        node = graph._nodes[rev]
        self.assertEqual(gdfo, node.gdfo)

    def test_children_ancestry1(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual([b"rev1"], graph.get_child_keys(NULL_REVISION))
        self.assertEqual([b"rev2a", b"rev2b"], sorted(graph.get_child_keys(b"rev1")))
        self.assertEqual([b"rev3"], graph.get_child_keys(b"rev2a"))
        self.assertEqual([b"rev4"], graph.get_child_keys(b"rev3"))
        self.assertEqual([b"rev4"], graph.get_child_keys(b"rev2b"))
        self.assertRaises(KeyError, graph.get_child_keys, b"not_in_graph")

    def test_parent_ancestry1(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual([NULL_REVISION], graph.get_parent_keys(b"rev1"))
        self.assertEqual([b"rev1"], graph.get_parent_keys(b"rev2a"))
        self.assertEqual([b"rev1"], graph.get_parent_keys(b"rev2b"))
        self.assertEqual([b"rev2a"], graph.get_parent_keys(b"rev3"))
        self.assertEqual([b"rev2b", b"rev3"], sorted(graph.get_parent_keys(b"rev4")))
        self.assertRaises(KeyError, graph.get_child_keys, b"not_in_graph")

    def test_parent_with_ghost(self):
        graph = self.make_known_graph(test_graph.with_ghost)
        self.assertEqual(None, graph.get_parent_keys(b"g"))

    def test_gdfo_ancestry_1(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertGDFO(graph, b"rev1", 2)
        self.assertGDFO(graph, b"rev2b", 3)
        self.assertGDFO(graph, b"rev2a", 3)
        self.assertGDFO(graph, b"rev3", 4)
        self.assertGDFO(graph, b"rev4", 5)

    def test_gdfo_feature_branch(self):
        graph = self.make_known_graph(test_graph.feature_branch)
        self.assertGDFO(graph, b"rev1", 2)
        self.assertGDFO(graph, b"rev2b", 3)
        self.assertGDFO(graph, b"rev3b", 4)

    def test_gdfo_extended_history_shortcut(self):
        graph = self.make_known_graph(test_graph.extended_history_shortcut)
        self.assertGDFO(graph, b"a", 2)
        self.assertGDFO(graph, b"b", 3)
        self.assertGDFO(graph, b"c", 4)
        self.assertGDFO(graph, b"d", 5)
        self.assertGDFO(graph, b"e", 6)
        self.assertGDFO(graph, b"f", 6)

    def test_gdfo_with_ghost(self):
        graph = self.make_known_graph(test_graph.with_ghost)
        self.assertGDFO(graph, b"f", 2)
        self.assertGDFO(graph, b"e", 3)
        self.assertGDFO(graph, b"g", 1)
        self.assertGDFO(graph, b"b", 4)
        self.assertGDFO(graph, b"d", 4)
        self.assertGDFO(graph, b"a", 5)
        self.assertGDFO(graph, b"c", 5)

    def test_add_existing_node(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        # Add a node that already exists with identical content
        # This is a 'no-op'
        self.assertGDFO(graph, b"rev4", 5)
        graph.add_node(b"rev4", [b"rev3", b"rev2b"])
        self.assertGDFO(graph, b"rev4", 5)
        # This also works if we use a tuple rather than a list
        graph.add_node(b"rev4", (b"rev3", b"rev2b"))

    def test_add_existing_node_mismatched_parents(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertRaises(ValueError, graph.add_node, b"rev4", [b"rev2b", b"rev3"])

    def test_add_node_with_ghost_parent(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        graph.add_node(b"rev5", [b"rev2b", b"revGhost"])
        self.assertGDFO(graph, b"rev5", 4)
        self.assertGDFO(graph, b"revGhost", 1)

    def test_add_new_root(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        graph.add_node(b"rev5", [])
        self.assertGDFO(graph, b"rev5", 1)

    def test_add_with_all_ghost_parents(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        graph.add_node(b"rev5", [b"ghost"])
        self.assertGDFO(graph, b"rev5", 2)
        self.assertGDFO(graph, b"ghost", 1)

    def test_gdfo_after_add_node(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual([], graph.get_child_keys(b"rev4"))
        graph.add_node(b"rev5", [b"rev4"])
        self.assertEqual([b"rev4"], graph.get_parent_keys(b"rev5"))
        self.assertEqual([b"rev5"], graph.get_child_keys(b"rev4"))
        self.assertEqual([], graph.get_child_keys(b"rev5"))
        self.assertGDFO(graph, b"rev5", 6)
        graph.add_node(b"rev6", [b"rev2b"])
        graph.add_node(b"rev7", [b"rev6"])
        graph.add_node(b"rev8", [b"rev7", b"rev5"])
        self.assertGDFO(graph, b"rev5", 6)
        self.assertGDFO(graph, b"rev6", 4)
        self.assertGDFO(graph, b"rev7", 5)
        self.assertGDFO(graph, b"rev8", 7)

    def test_fill_in_ghost(self):
        graph = self.make_known_graph(test_graph.with_ghost)
        # Add in a couple nodes and then fill in the 'ghost' so that it should
        # cause renumbering of children nodes
        graph.add_node(b"x", [])
        graph.add_node(b"y", [b"x"])
        graph.add_node(b"z", [b"y"])
        graph.add_node(b"g", [b"z"])
        self.assertGDFO(graph, b"f", 2)
        self.assertGDFO(graph, b"e", 3)
        self.assertGDFO(graph, b"x", 1)
        self.assertGDFO(graph, b"y", 2)
        self.assertGDFO(graph, b"z", 3)
        self.assertGDFO(graph, b"g", 4)
        self.assertGDFO(graph, b"b", 4)
        self.assertGDFO(graph, b"d", 5)
        self.assertGDFO(graph, b"a", 5)
        self.assertGDFO(graph, b"c", 6)


class TestKnownGraphHeads(TestCaseWithKnownGraph):
    scenarios = caching_scenarios() + non_caching_scenarios()
    do_cache = None  # Set by load_tests

    def test_heads_null(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual({b"null:"}, graph.heads([b"null:"]))
        self.assertEqual({b"rev1"}, graph.heads([b"null:", b"rev1"]))
        self.assertEqual({b"rev1"}, graph.heads([b"rev1", b"null:"]))
        self.assertEqual({b"rev1"}, graph.heads({b"rev1", b"null:"}))
        self.assertEqual({b"rev1"}, graph.heads((b"rev1", b"null:")))

    def test_heads_one(self):
        # A single node will always be a head
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual({b"null:"}, graph.heads([b"null:"]))
        self.assertEqual({b"rev1"}, graph.heads([b"rev1"]))
        self.assertEqual({b"rev2a"}, graph.heads([b"rev2a"]))
        self.assertEqual({b"rev2b"}, graph.heads([b"rev2b"]))
        self.assertEqual({b"rev3"}, graph.heads([b"rev3"]))
        self.assertEqual({b"rev4"}, graph.heads([b"rev4"]))

    def test_heads_single(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual({b"rev4"}, graph.heads([b"null:", b"rev4"]))
        self.assertEqual({b"rev2a"}, graph.heads([b"rev1", b"rev2a"]))
        self.assertEqual({b"rev2b"}, graph.heads([b"rev1", b"rev2b"]))
        self.assertEqual({b"rev3"}, graph.heads([b"rev1", b"rev3"]))
        self.assertEqual({b"rev3"}, graph.heads([b"rev3", b"rev2a"]))
        self.assertEqual({b"rev4"}, graph.heads([b"rev1", b"rev4"]))
        self.assertEqual({b"rev4"}, graph.heads([b"rev2a", b"rev4"]))
        self.assertEqual({b"rev4"}, graph.heads([b"rev2b", b"rev4"]))
        self.assertEqual({b"rev4"}, graph.heads([b"rev3", b"rev4"]))

    def test_heads_two_heads(self):
        graph = self.make_known_graph(test_graph.ancestry_1)
        self.assertEqual({b"rev2a", b"rev2b"}, graph.heads([b"rev2a", b"rev2b"]))
        self.assertEqual({b"rev3", b"rev2b"}, graph.heads([b"rev3", b"rev2b"]))

    def test_heads_criss_cross(self):
        graph = self.make_known_graph(test_graph.criss_cross)
        self.assertEqual({b"rev2a"}, graph.heads([b"rev2a", b"rev1"]))
        self.assertEqual({b"rev2b"}, graph.heads([b"rev2b", b"rev1"]))
        self.assertEqual({b"rev3a"}, graph.heads([b"rev3a", b"rev1"]))
        self.assertEqual({b"rev3b"}, graph.heads([b"rev3b", b"rev1"]))
        self.assertEqual({b"rev2a", b"rev2b"}, graph.heads([b"rev2a", b"rev2b"]))
        self.assertEqual({b"rev3a"}, graph.heads([b"rev3a", b"rev2a"]))
        self.assertEqual({b"rev3a"}, graph.heads([b"rev3a", b"rev2b"]))
        self.assertEqual({b"rev3a"}, graph.heads([b"rev3a", b"rev2a", b"rev2b"]))
        self.assertEqual({b"rev3b"}, graph.heads([b"rev3b", b"rev2a"]))
        self.assertEqual({b"rev3b"}, graph.heads([b"rev3b", b"rev2b"]))
        self.assertEqual({b"rev3b"}, graph.heads([b"rev3b", b"rev2a", b"rev2b"]))
        self.assertEqual({b"rev3a", b"rev3b"}, graph.heads([b"rev3a", b"rev3b"]))
        self.assertEqual(
            {b"rev3a", b"rev3b"}, graph.heads([b"rev3a", b"rev3b", b"rev2a", b"rev2b"])
        )

    def test_heads_shortcut(self):
        graph = self.make_known_graph(test_graph.history_shortcut)
        self.assertEqual(
            {b"rev2a", b"rev2b", b"rev2c"}, graph.heads([b"rev2a", b"rev2b", b"rev2c"])
        )
        self.assertEqual({b"rev3a", b"rev3b"}, graph.heads([b"rev3a", b"rev3b"]))
        self.assertEqual(
            {b"rev3a", b"rev3b"}, graph.heads([b"rev2a", b"rev3a", b"rev3b"])
        )
        self.assertEqual({b"rev2a", b"rev3b"}, graph.heads([b"rev2a", b"rev3b"]))
        self.assertEqual({b"rev2c", b"rev3a"}, graph.heads([b"rev2c", b"rev3a"]))

    def test_heads_linear(self):
        graph = self.make_known_graph(test_graph.racing_shortcuts)
        self.assertEqual({b"w"}, graph.heads([b"w", b"s"]))
        self.assertEqual({b"z"}, graph.heads([b"w", b"s", b"z"]))
        self.assertEqual({b"w", b"q"}, graph.heads([b"w", b"s", b"q"]))
        self.assertEqual({b"z"}, graph.heads([b"s", b"z"]))

    def test_heads_alt_merge(self):
        graph = self.make_known_graph(alt_merge)
        self.assertEqual({b"c"}, graph.heads([b"a", b"c"]))

    def test_heads_with_ghost(self):
        graph = self.make_known_graph(test_graph.with_ghost)
        self.assertEqual({b"e", b"g"}, graph.heads([b"e", b"g"]))
        self.assertEqual({b"a", b"c"}, graph.heads([b"a", b"c"]))
        self.assertEqual({b"a", b"g"}, graph.heads([b"a", b"g"]))
        self.assertEqual({b"f", b"g"}, graph.heads([b"f", b"g"]))
        self.assertEqual({b"c"}, graph.heads([b"c", b"g"]))
        self.assertEqual({b"c"}, graph.heads([b"c", b"b", b"d", b"g"]))
        self.assertEqual({b"a", b"c"}, graph.heads([b"a", b"c", b"e", b"g"]))
        self.assertEqual({b"a", b"c"}, graph.heads([b"a", b"c", b"f"]))

    def test_filling_in_ghosts_resets_head_cache(self):
        graph = self.make_known_graph(test_graph.with_ghost)
        self.assertEqual({b"e", b"g"}, graph.heads([b"e", b"g"]))
        # 'g' is filled in, and decends from 'e', so the heads result is now
        # different
        graph.add_node(b"g", [b"e"])
        self.assertEqual({b"g"}, graph.heads([b"e", b"g"]))


class TestKnownGraphTopoSort(TestCaseWithKnownGraph):
    def assertTopoSortOrder(self, ancestry):
        """Check topo_sort and iter_topo_order is genuinely topological order.

        For every child in the graph, check if it comes after all of it's
        parents.
        """
        graph = self.make_known_graph(ancestry)
        sort_result = graph.topo_sort()
        # We should have an entry in sort_result for every entry present in the
        # graph.
        self.assertEqual(len(ancestry), len(sort_result))
        node_idx = {node: idx for idx, node in enumerate(sort_result)}
        for node in sort_result:
            parents = ancestry[node]
            for parent in parents:
                if parent not in ancestry:
                    # ghost
                    continue
                if node_idx[node] <= node_idx[parent]:
                    self.fail(
                        "parent {} must come before child {}:\n{}".format(
                            parent, node, sort_result
                        )
                    )

    def test_topo_sort_empty(self):
        """TopoSort empty list."""
        self.assertTopoSortOrder({})

    def test_topo_sort_easy(self):
        """TopoSort list with one node."""
        self.assertTopoSortOrder({0: []})

    def test_topo_sort_cycle(self):
        """TopoSort traps graph with cycles."""
        g = self.make_known_graph({0: [1], 1: [0]})
        self.assertRaises(errors.GraphCycleError, g.topo_sort)

    def test_topo_sort_cycle_2(self):
        """TopoSort traps graph with longer cycle."""
        g = self.make_known_graph({0: [1], 1: [2], 2: [0]})
        self.assertRaises(errors.GraphCycleError, g.topo_sort)

    def test_topo_sort_cycle_with_tail(self):
        """TopoSort traps graph with longer cycle."""
        g = self.make_known_graph({0: [1], 1: [2], 2: [3, 4], 3: [0], 4: []})
        self.assertRaises(errors.GraphCycleError, g.topo_sort)

    def test_topo_sort_1(self):
        """TopoSort simple nontrivial graph."""
        self.assertTopoSortOrder({0: [3], 1: [4], 2: [1, 4], 3: [], 4: [0, 3]})

    def test_topo_sort_partial(self):
        """Topological sort with partial ordering.

        Multiple correct orderings are possible, so test for
        correctness, not for exact match on the resulting list.
        """
        self.assertTopoSortOrder(
            {
                0: [],
                1: [0],
                2: [0],
                3: [0],
                4: [1, 2, 3],
                5: [1, 2],
                6: [1, 2],
                7: [2, 3],
                8: [0, 1, 4, 5, 6],
            }
        )

    def test_topo_sort_ghost_parent(self):
        """Sort nodes, but don't include some parents in the output."""
        self.assertTopoSortOrder({0: [1], 1: [2]})


class TestKnownGraphMergeSort(TestCaseWithKnownGraph):
    def assertSortAndIterate(self, ancestry, branch_tip, result_list):
        """Check that merge based sorting and iter_topo_order on graph works."""
        graph = self.make_known_graph(ancestry)
        value = graph.merge_sort(branch_tip)
        value = [(n.key, n.merge_depth, n.revno, n.end_of_merge) for n in value]
        if result_list != value:
            self.assertEqualDiff(pprint.pformat(result_list), pprint.pformat(value))

    def test_merge_sort_empty(self):
        # sorting of an emptygraph does not error
        self.assertSortAndIterate({}, None, [])
        self.assertSortAndIterate({}, NULL_REVISION, [])
        self.assertSortAndIterate({}, (NULL_REVISION,), [])

    def test_merge_sort_not_empty_no_tip(self):
        # merge sorting of a branch starting with None should result
        # in an empty list: no revisions are dragged in.
        self.assertSortAndIterate({0: []}, None, [])
        self.assertSortAndIterate({0: []}, NULL_REVISION, [])
        self.assertSortAndIterate({0: []}, (NULL_REVISION,), [])

    def test_merge_sort_one_revision(self):
        # sorting with one revision as the tip returns the correct fields:
        # sequence - 0, revision id, merge depth - 0, end_of_merge
        self.assertSortAndIterate({"id": []}, "id", [("id", 0, (1,), True)])

    def test_sequence_numbers_increase_no_merges(self):
        # emit a few revisions with no merges to check the sequence
        # numbering works in trivial cases
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["B"]},
            "C",
            [
                ("C", 0, (3,), False),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_sequence_numbers_increase_with_merges(self):
        # test that sequence numbers increase across merges
        self.assertSortAndIterate(
            {"A": [], "B": ["A"], "C": ["A", "B"]},
            "C",
            [
                ("C", 0, (2,), False),
                ("B", 1, (1, 1, 1), True),
                ("A", 0, (1,), True),
            ],
        )

    def test_merge_sort_race(self):
        # A
        # |
        # B-.
        # |\ \
        # | | C
        # | |/
        # | D
        # |/
        # F
        graph = {
            "A": [],
            "B": ["A"],
            "C": ["B"],
            "D": ["B", "C"],
            "F": ["B", "D"],
        }
        self.assertSortAndIterate(
            graph,
            "F",
            [
                ("F", 0, (3,), False),
                ("D", 1, (2, 2, 1), False),
                ("C", 2, (2, 1, 1), True),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )
        # A
        # |
        # B-.
        # |\ \
        # | X C
        # | |/
        # | D
        # |/
        # F
        graph = {
            "A": [],
            "B": ["A"],
            "C": ["B"],
            "X": ["B"],
            "D": ["X", "C"],
            "F": ["B", "D"],
        }
        self.assertSortAndIterate(
            graph,
            "F",
            [
                ("F", 0, (3,), False),
                ("D", 1, (2, 1, 2), False),
                ("C", 2, (2, 2, 1), True),
                ("X", 1, (2, 1, 1), True),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_merge_depth_with_nested_merges(self):
        # the merge depth marker should reflect the depth of the revision
        # in terms of merges out from the mainline
        # revid, depth, parents:
        #  A 0   [D, B]
        #  B  1  [C, F]
        #  C  1  [H]
        #  D 0   [H, E]
        #  E  1  [G, F]
        #  F   2 [G]
        #  G  1  [H]
        #  H 0
        self.assertSortAndIterate(
            {
                "A": ["D", "B"],
                "B": ["C", "F"],
                "C": ["H"],
                "D": ["H", "E"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            },
            "A",
            [
                ("A", 0, (3,), False),
                ("B", 1, (1, 3, 2), False),
                ("C", 1, (1, 3, 1), True),
                ("D", 0, (2,), False),
                ("E", 1, (1, 1, 2), False),
                ("F", 2, (1, 2, 1), True),
                ("G", 1, (1, 1, 1), True),
                ("H", 0, (1,), True),
            ],
        )

    def test_dotted_revnos_with_simple_merges(self):
        # A         1
        # |\
        # B C       2, 1.1.1
        # | |\
        # D E F     3, 1.1.2, 1.2.1
        # |/ /|
        # G H I     4, 1.2.2, 1.3.1
        # |/ /
        # J K       5, 1.3.2
        # |/
        # L         6
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["A"],
                "D": ["B"],
                "E": ["C"],
                "F": ["C"],
                "G": ["D", "E"],
                "H": ["F"],
                "I": ["F"],
                "J": ["G", "H"],
                "K": ["I"],
                "L": ["J", "K"],
            },
            "L",
            [
                ("L", 0, (6,), False),
                ("K", 1, (1, 3, 2), False),
                ("I", 1, (1, 3, 1), True),
                ("J", 0, (5,), False),
                ("H", 1, (1, 2, 2), False),
                ("F", 1, (1, 2, 1), True),
                ("G", 0, (4,), False),
                ("E", 1, (1, 1, 2), False),
                ("C", 1, (1, 1, 1), True),
                ("D", 0, (3,), False),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )
        # Adding a shortcut from the first revision should not change any of
        # the existing numbers
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["A"],
                "D": ["B"],
                "E": ["C"],
                "F": ["C"],
                "G": ["D", "E"],
                "H": ["F"],
                "I": ["F"],
                "J": ["G", "H"],
                "K": ["I"],
                "L": ["J", "K"],
                "M": ["A"],
                "N": ["L", "M"],
            },
            "N",
            [
                ("N", 0, (7,), False),
                ("M", 1, (1, 4, 1), True),
                ("L", 0, (6,), False),
                ("K", 1, (1, 3, 2), False),
                ("I", 1, (1, 3, 1), True),
                ("J", 0, (5,), False),
                ("H", 1, (1, 2, 2), False),
                ("F", 1, (1, 2, 1), True),
                ("G", 0, (4,), False),
                ("E", 1, (1, 1, 2), False),
                ("C", 1, (1, 1, 1), True),
                ("D", 0, (3,), False),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_end_of_merge_not_last_revision_in_branch(self):
        # within a branch only the last revision gets an
        # end of merge marker.
        self.assertSortAndIterate(
            {
                "A": ["B"],
                "B": [],
            },
            "A",
            [("A", 0, (2,), False), ("B", 0, (1,), True)],
        )

    def test_end_of_merge_multiple_revisions_merged_at_once(self):
        # when multiple branches are merged at once, both of their
        # branch-endpoints should be listed as end-of-merge.
        # Also, the order of the multiple merges should be
        # left-right shown top to bottom.
        # * means end of merge
        # A 0    [H, B, E]
        # B  1   [D, C]
        # C   2  [D]       *
        # D  1   [H]       *
        # E  1   [G, F]
        # F   2  [G]       *
        # G  1   [H]       *
        # H 0    []        *
        self.assertSortAndIterate(
            {
                "A": ["H", "B", "E"],
                "B": ["D", "C"],
                "C": ["D"],
                "D": ["H"],
                "E": ["G", "F"],
                "F": ["G"],
                "G": ["H"],
                "H": [],
            },
            "A",
            [
                ("A", 0, (2,), False),
                ("B", 1, (1, 3, 2), False),
                ("C", 2, (1, 4, 1), True),
                ("D", 1, (1, 3, 1), True),
                ("E", 1, (1, 1, 2), False),
                ("F", 2, (1, 2, 1), True),
                ("G", 1, (1, 1, 1), True),
                ("H", 0, (1,), True),
            ],
        )

    def test_parallel_root_sequence_numbers_increase_with_merges(self):
        """When there are parallel roots, check their revnos."""
        self.assertSortAndIterate(
            {"A": [], "B": [], "C": ["A", "B"]},
            "C",
            [
                ("C", 0, (2,), False),
                ("B", 1, (0, 1, 1), True),
                ("A", 0, (1,), True),
            ],
        )

    def test_revnos_are_globally_assigned(self):
        """Revnos are assigned according to the revision they derive from."""
        # in this test we setup a number of branches that all derive from
        # the first revision, and then merge them one at a time, which
        # should give the revisions as they merge numbers still deriving from
        # the revision were based on.
        # merge 3: J: ['G', 'I']
        # branch 3:
        #  I: ['H']
        #  H: ['A']
        # merge 2: G: ['D', 'F']
        # branch 2:
        #  F: ['E']
        #  E: ['A']
        # merge 1: D: ['A', 'C']
        # branch 1:
        #  C: ['B']
        #  B: ['A']
        # root: A: []
        self.assertSortAndIterate(
            {
                "J": ["G", "I"],
                "I": [
                    "H",
                ],
                "H": ["A"],
                "G": ["D", "F"],
                "F": ["E"],
                "E": ["A"],
                "D": ["A", "C"],
                "C": ["B"],
                "B": ["A"],
                "A": [],
            },
            "J",
            [
                ("J", 0, (4,), False),
                ("I", 1, (1, 3, 2), False),
                ("H", 1, (1, 3, 1), True),
                ("G", 0, (3,), False),
                ("F", 1, (1, 2, 2), False),
                ("E", 1, (1, 2, 1), True),
                ("D", 0, (2,), False),
                ("C", 1, (1, 1, 2), False),
                ("B", 1, (1, 1, 1), True),
                ("A", 0, (1,), True),
            ],
        )

    def test_roots_and_sub_branches_versus_ghosts(self):
        """Extra roots and their mini branches use the same numbering.

        All of them use the 0-node numbering.
        """
        #       A D   K
        #       | |\  |\
        #       B E F L M
        #       | |/  |/
        #       C G   N
        #       |/    |\
        #       H I   O P
        #       |/    |/
        #       J     Q
        #       |.---'
        #       R
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["B"],
                "D": [],
                "E": ["D"],
                "F": ["D"],
                "G": ["E", "F"],
                "H": ["C", "G"],
                "I": [],
                "J": ["H", "I"],
                "K": [],
                "L": ["K"],
                "M": ["K"],
                "N": ["L", "M"],
                "O": ["N"],
                "P": ["N"],
                "Q": ["O", "P"],
                "R": ["J", "Q"],
            },
            "R",
            [
                ("R", 0, (6,), False),
                ("Q", 1, (0, 4, 5), False),
                ("P", 2, (0, 6, 1), True),
                ("O", 1, (0, 4, 4), False),
                ("N", 1, (0, 4, 3), False),
                ("M", 2, (0, 5, 1), True),
                ("L", 1, (0, 4, 2), False),
                ("K", 1, (0, 4, 1), True),
                ("J", 0, (5,), False),
                ("I", 1, (0, 3, 1), True),
                ("H", 0, (4,), False),
                ("G", 1, (0, 1, 3), False),
                ("F", 2, (0, 2, 1), True),
                ("E", 1, (0, 1, 2), False),
                ("D", 1, (0, 1, 1), True),
                ("C", 0, (3,), False),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_ghost(self):
        # merge_sort should be able to ignore ghosts
        # A
        # |
        # B ghost
        # |/
        # C
        self.assertSortAndIterate(
            {
                "A": [],
                "B": ["A"],
                "C": ["B", "ghost"],
            },
            "C",
            [
                ("C", 0, (3,), False),
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_lefthand_ghost(self):
        # ghost
        #  |
        #  A
        #  |
        #  B
        self.assertSortAndIterate(
            {
                "A": ["ghost"],
                "B": ["A"],
            },
            "B",
            [
                ("B", 0, (2,), False),
                ("A", 0, (1,), True),
            ],
        )

    def test_graph_cycle(self):
        # merge_sort should fail with a simple error when a graph cycle is
        # encountered.
        #
        # A
        # |,-.
        # B  |
        # |  |
        # C  ^
        # |  |
        # D  |
        # |'-'
        # E
        self.assertRaises(
            errors.GraphCycleError,
            self.assertSortAndIterate,
            {
                "A": [],
                "B": ["D"],
                "C": ["B"],
                "D": ["C"],
                "E": ["D"],
            },
            "E",
            [],
        )


class TestKnownGraphStableReverseTopoSort(TestCaseWithKnownGraph):
    """Test the sort order returned by gc_sort."""

    def assertSorted(self, expected, parent_map):
        graph = self.make_known_graph(parent_map)
        value = graph.gc_sort()
        if expected != value:
            self.assertEqualDiff(pprint.pformat(expected), pprint.pformat(value))

    def test_empty(self):
        self.assertSorted([], {})

    def test_single(self):
        self.assertSorted(["a"], {"a": ()})
        self.assertSorted([("a",)], {("a",): ()})
        self.assertSorted([("F", "a")], {("F", "a"): ()})

    def test_linear(self):
        self.assertSorted(["c", "b", "a"], {"a": (), "b": ("a",), "c": ("b",)})
        self.assertSorted(
            [("c",), ("b",), ("a",)], {("a",): (), ("b",): (("a",),), ("c",): (("b",),)}
        )
        self.assertSorted(
            [("F", "c"), ("F", "b"), ("F", "a")],
            {("F", "a"): (), ("F", "b"): (("F", "a"),), ("F", "c"): (("F", "b"),)},
        )

    def test_mixed_ancestries(self):
        # Each prefix should be sorted separately
        self.assertSorted(
            [
                ("F", "c"),
                ("F", "b"),
                ("F", "a"),
                ("G", "c"),
                ("G", "b"),
                ("G", "a"),
                ("Q", "c"),
                ("Q", "b"),
                ("Q", "a"),
            ],
            {
                ("F", "a"): (),
                ("F", "b"): (("F", "a"),),
                ("F", "c"): (("F", "b"),),
                ("G", "a"): (),
                ("G", "b"): (("G", "a"),),
                ("G", "c"): (("G", "b"),),
                ("Q", "a"): (),
                ("Q", "b"): (("Q", "a"),),
                ("Q", "c"): (("Q", "b"),),
            },
        )

    def test_stable_sorting(self):
        # the sort order should be stable even when extra nodes are added
        self.assertSorted(["b", "c", "a"], {"a": (), "b": ("a",), "c": ("a",)})
        self.assertSorted(
            ["b", "c", "d", "a"], {"a": (), "b": ("a",), "c": ("a",), "d": ("a",)}
        )
        self.assertSorted(
            ["b", "c", "d", "a"], {"a": (), "b": ("a",), "c": ("a",), "d": ("a",)}
        )
        self.assertSorted(
            ["Z", "b", "c", "d", "a"],
            {"a": (), "b": ("a",), "c": ("a",), "d": ("a",), "Z": ("a",)},
        )
        self.assertSorted(
            ["e", "b", "c", "f", "Z", "d", "a"],
            {
                "a": (),
                "b": ("a",),
                "c": ("a",),
                "d": ("a",),
                "Z": ("a",),
                "e": ("b", "c", "d"),
                "f": ("d", "Z"),
            },
        )

    def test_skip_ghost(self):
        self.assertSorted(["b", "c", "a"], {"a": (), "b": ("a", "ghost"), "c": ("a",)})

    def test_skip_mainline_ghost(self):
        self.assertSorted(["b", "c", "a"], {"a": (), "b": ("ghost", "a"), "c": ("a",)})



# Copyright (C) 2006, 2007, 2009-2013, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz sign-my-commits."""

from breezy import gpg, tests


class SignMyCommits(tests.TestCaseWithTransport):
    def monkey_patch_gpg(self):
        """Monkey patch the gpg signing strategy to be a loopback.

        This also registers the cleanup, so that we will revert to
        the original gpg strategy when done.
        """
        # monkey patch gpg signing mechanism
        self.overrideAttr(gpg, "GPGStrategy", gpg.LoopbackGPGStrategy)

    def setup_tree(self, location="."):
        wt = self.make_branch_and_tree(location)
        wt.commit("base A", allow_pointless=True, rev_id=b"A")
        wt.commit("base B", allow_pointless=True, rev_id=b"B")
        wt.commit("base C", allow_pointless=True, rev_id=b"C")
        wt.commit(
            "base D",
            allow_pointless=True,
            rev_id=b"D",
            committer="Alternate <alt@foo.com>",
        )
        wt.add_parent_tree_id(b"aghost")
        wt.commit("base E", allow_pointless=True, rev_id=b"E")
        return wt

    def assertUnsigned(self, repo, revision_id):
        """Assert that revision_id is not signed in repo."""
        self.assertFalse(repo.has_signature_for_revision_id(revision_id))

    def assertSigned(self, repo, revision_id):
        """Assert that revision_id is signed in repo."""
        self.assertTrue(repo.has_signature_for_revision_id(revision_id))

    def test_sign_my_commits(self):
        # Test re signing of data.
        wt = self.setup_tree()
        repo = wt.branch.repository

        self.monkey_patch_gpg()

        self.assertUnsigned(repo, b"A")
        self.assertUnsigned(repo, b"B")
        self.assertUnsigned(repo, b"C")
        self.assertUnsigned(repo, b"D")

        self.run_bzr("sign-my-commits")

        self.assertSigned(repo, b"A")
        self.assertSigned(repo, b"B")
        self.assertSigned(repo, b"C")
        self.assertUnsigned(repo, b"D")

    def test_sign_my_commits_location(self):
        wt = self.setup_tree("other")
        repo = wt.branch.repository

        self.monkey_patch_gpg()

        self.run_bzr("sign-my-commits other")

        self.assertSigned(repo, b"A")
        self.assertSigned(repo, b"B")
        self.assertSigned(repo, b"C")
        self.assertUnsigned(repo, b"D")

    def test_sign_diff_committer(self):
        wt = self.setup_tree()
        repo = wt.branch.repository

        self.monkey_patch_gpg()

        self.run_bzr(["sign-my-commits", ".", "Alternate <alt@foo.com>"])

        self.assertUnsigned(repo, b"A")
        self.assertUnsigned(repo, b"B")
        self.assertUnsigned(repo, b"C")
        self.assertSigned(repo, b"D")

    def test_sign_dry_run(self):
        wt = self.setup_tree()
        repo = wt.branch.repository

        self.monkey_patch_gpg()

        out = self.run_bzr("sign-my-commits --dry-run")[0]

        outlines = out.splitlines()
        self.assertEqual(5, len(outlines))
        self.assertEqual("Signed 4 revisions.", outlines[-1])
        self.assertUnsigned(repo, b"A")
        self.assertUnsigned(repo, b"B")
        self.assertUnsigned(repo, b"C")
        self.assertUnsigned(repo, b"D")
        self.assertUnsigned(repo, b"E")



# Copyright (C) 2005, 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import tests
from breezy.tests import features


class TestLSProf(tests.TestCaseInTempDir):
    _test_needs_features = [features.lsprof_feature]

    def test_file(self):
        out, err = self.run_bzr("--lsprof-file output.callgrind rocks")
        self.assertNotContainsRe(out, "Profile data written to")
        self.assertContainsRe(err, "Profile data written to")

    def test_stdout(self):
        out, err = self.run_bzr("--lsprof rocks")
        self.assertContainsRe(out, "CallCount")
        self.assertNotContainsRe(err, "Profile data written to")



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the 'brz dump-btree' command."""

from breezy import tests
from breezy.bzr import btree_index
from breezy.tests import http_server


class TestDumpBtree(tests.TestCaseWithTransport):
    def create_sample_btree_index(self):
        builder = btree_index.BTreeBuilder(reference_lists=1, key_elements=2)
        builder.add_node((b"test", b"key1"), b"value", (((b"ref", b"entry"),),))
        builder.add_node((b"test", b"key2"), b"value2", (((b"ref", b"entry2"),),))
        builder.add_node((b"test2", b"key3"), b"value3", (((b"ref", b"entry3"),),))
        out_f = builder.finish()
        try:
            self.build_tree_contents([("test.btree", out_f.read())])
        finally:
            out_f.close()

    def test_dump_btree_smoke(self):
        self.create_sample_btree_index()
        out, err = self.run_bzr("dump-btree test.btree")
        self.assertEqualDiff(
            "(('test', 'key1'), 'value', ((('ref', 'entry'),),))\n"
            "(('test', 'key2'), 'value2', ((('ref', 'entry2'),),))\n"
            "(('test2', 'key3'), 'value3', ((('ref', 'entry3'),),))\n",
            out,
        )

    def test_dump_btree_http_smoke(self):
        self.transport_readonly_server = http_server.HttpServer
        self.create_sample_btree_index()
        url = self.get_readonly_url("test.btree")
        out, err = self.run_bzr(["dump-btree", url])
        self.assertEqualDiff(
            "(('test', 'key1'), 'value', ((('ref', 'entry'),),))\n"
            "(('test', 'key2'), 'value2', ((('ref', 'entry2'),),))\n"
            "(('test2', 'key3'), 'value3', ((('ref', 'entry3'),),))\n",
            out,
        )

    def test_dump_btree_raw_smoke(self):
        self.create_sample_btree_index()
        out, err = self.run_bzr("dump-btree test.btree --raw")
        self.assertEqualDiff(
            "Root node:\n"
            "B+Tree Graph Index 2\n"
            "node_ref_lists=1\n"
            "key_elements=2\n"
            "len=3\n"
            "row_lengths=1\n"
            "\n"
            "Page 0\n"
            "type=leaf\n"
            "test\0key1\0ref\0entry\0value\n"
            "test\0key2\0ref\0entry2\0value2\n"
            "test2\0key3\0ref\0entry3\0value3\n"
            "\n",
            out,
        )

    def test_dump_btree_no_refs_smoke(self):
        # A BTree index with no ref lists (such as *.cix) can be dumped without
        # errors.
        builder = btree_index.BTreeBuilder(reference_lists=0, key_elements=2)
        builder.add_node((b"test", b"key1"), b"value")
        out_f = builder.finish()
        try:
            self.build_tree_contents([("test.btree", out_f.read())])
        finally:
            out_f.close()
        out, err = self.run_bzr("dump-btree test.btree")

    def create_sample_empty_btree_index(self):
        builder = btree_index.BTreeBuilder(reference_lists=1, key_elements=2)
        out_f = builder.finish()
        try:
            self.build_tree_contents([("test.btree", out_f.read())])
        finally:
            out_f.close()

    def test_dump_empty_btree_smoke(self):
        self.create_sample_empty_btree_index()
        out, err = self.run_bzr("dump-btree test.btree")
        self.assertEqualDiff("", out)

    def test_dump_empty_btree_http_smoke(self):
        self.transport_readonly_server = http_server.HttpServer
        self.create_sample_empty_btree_index()
        url = self.get_readonly_url("test.btree")
        out, err = self.run_bzr(["dump-btree", url])
        self.assertEqualDiff("", out)

    def test_dump_empty_btree_raw_smoke(self):
        self.create_sample_empty_btree_index()
        out, err = self.run_bzr("dump-btree test.btree --raw")
        self.assertEqualDiff(
            "Root node:\n"
            "B+Tree Graph Index 2\n"
            "node_ref_lists=1\n"
            "key_elements=2\n"
            "len=0\n"
            "row_lengths=\n"
            "\n"
            "Page 0\n"
            "(empty)\n",
            out,
        )



# Copyright (C) 2017 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import TYPE_CHECKING

from .. import config, controldir, errors, pyutils, registry
from .. import transport as _mod_transport
from ..branch import format_registry as branch_format_registry
from ..repository import format_registry as repository_format_registry
from ..workingtree import format_registry as workingtree_format_registry

if TYPE_CHECKING:
    from .bzrdir import BzrDirFormat


class LineEndingError(errors.BzrError):
    _fmt = (
        "Line ending corrupted for file: %(file)s; "
        "Maybe your files got corrupted in transport?"
    )

    def __init__(self, file):
        self.file = file


class BzrProber(controldir.Prober):
    """Prober for formats that use a .bzr/ control directory."""

    formats = registry.FormatRegistry["BzrDirFormat"](
        controldir.network_format_registry
    )
    """The known .bzr formats."""

    @classmethod
    def priority(klass, transport):
        return 10

    @classmethod
    def probe_transport(klass, transport):
        """Return the .bzrdir style format present in a directory."""
        try:
            format_string = transport.get_bytes(".bzr/branch-format")
        except _mod_transport.NoSuchFile as e:
            raise errors.NotBranchError(path=transport.base) from e
        except errors.BadHttpRequest as e:
            if e.reason == "no such method: .bzr":
                # hgweb
                raise errors.NotBranchError(path=transport.base) from e
            raise

        try:
            first_line = format_string[: format_string.index(b"\n") + 1]
        except ValueError:
            first_line = format_string
        if first_line.startswith(b"<!DOCTYPE") or first_line.startswith(b"<html"):
            raise errors.NotBranchError(
                path=transport.base, detail="format file looks like HTML"
            )
        try:
            cls = klass.formats.get(first_line)
        except KeyError as e:
            if first_line.endswith(b"\r\n"):
                raise LineEndingError(file=".bzr/branch-format") from e
            else:
                raise errors.UnknownFormatError(format=first_line, kind="bzrdir") from e
        return cls.from_string(format_string)

    @classmethod
    def known_formats(cls):
        result = []
        for _name, format in cls.formats.items():
            if callable(format):
                format = format()
            result.append(format)
        return result


controldir.ControlDirFormat.register_prober(BzrProber)


class RemoteBzrProber(controldir.Prober):
    """Prober for remote servers that provide a Bazaar smart server."""

    @classmethod
    def priority(klass, transport):
        return -10

    @classmethod
    def probe_transport(klass, transport):
        """Return a RemoteBzrDirFormat object if it looks possible."""
        try:
            medium = transport.get_smart_medium()
        except (
            NotImplementedError,
            AttributeError,
            errors.TransportNotPossible,
            errors.NoSmartMedium,
            errors.SmartProtocolError,
        ) as e:
            # no smart server, so not a branch for this format type.
            raise errors.NotBranchError(path=transport.base) from e
        else:
            # Decline to open it if the server doesn't support our required
            # version (3) so that the VFS-based transport will do it.
            if medium.should_probe():
                try:
                    server_version = medium.protocol_version()
                except errors.SmartProtocolError as e:
                    # Apparently there's no usable smart server there, even though
                    # the medium supports the smart protocol.
                    raise errors.NotBranchError(path=transport.base) from e
                if server_version != "2":
                    raise errors.NotBranchError(path=transport.base)
            from .remote import RemoteBzrDirFormat

            return RemoteBzrDirFormat()

    @classmethod
    def known_formats(cls):
        from .remote import RemoteBzrDirFormat

        return [RemoteBzrDirFormat()]


controldir.ControlDirFormat.register_prober(RemoteBzrProber)

# Register bzr formats
BzrProber.formats.register_lazy(
    b"Bazaar-NG meta directory, format 1\n", __name__ + ".bzrdir", "BzrDirMetaFormat1"
)
BzrProber.formats.register_lazy(
    b"Bazaar meta directory, format 1 (with colocated branches)\n",
    __name__ + ".bzrdir",
    "BzrDirMetaFormat1Colo",
)


def register_metadir(
    registry,
    key,
    repository_format,
    help,
    native=True,
    deprecated=False,
    branch_format=None,
    tree_format=None,
    hidden=False,
    experimental=False,
    bzrdir_format=None,
):
    """Register a metadir subformat.

    These all use a meta bzrdir, but can be parameterized by the
    Repository/Branch/WorkingTreeformats.

    :param repository_format: The fully-qualified repository format class
        name as a string.
    :param branch_format: Fully-qualified branch format class name as
        a string.
    :param tree_format: Fully-qualified tree format class name as
        a string.
    """
    if bzrdir_format is None:
        bzrdir_format = "breezy.bzr.bzrdir.BzrDirMetaFormat1"
    # This should be expanded to support setting WorkingTree and Branch
    # formats, once the API supports that.

    def _load(full_name):
        mod_name, factory_name = full_name.rsplit(".", 1)
        try:
            factory = pyutils.get_named_object(mod_name, factory_name)
        except ImportError as e:
            raise ImportError(f"failed to load {full_name}: {e}") from e
        except AttributeError as e:
            raise AttributeError(
                "no factory {} in module {!r}".format(full_name, sys.modules[mod_name])
            ) from e
        return factory()

    def helper():
        bd = _load(bzrdir_format)
        if branch_format is not None:
            bd.set_branch_format(_load(branch_format))
        if tree_format is not None:
            bd.workingtree_format = _load(tree_format)
        if repository_format is not None:
            bd.repository_format = _load(repository_format)
        return bd

    registry.register(key, helper, help, native, deprecated, hidden, experimental)


# formats which have no format string are not discoverable
# and not independently creatable, so are not registered.
branch_format_registry.register_lazy(
    b"Bazaar-NG branch format 5\n", "breezy.bzr.fullhistory", "BzrBranchFormat5"
)
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 6 (bzr 0.15)\n", "breezy.bzr.branch", "BzrBranchFormat6"
)
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 7 (needs bzr 1.6)\n", "breezy.bzr.branch", "BzrBranchFormat7"
)
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 8 (needs bzr 1.15)\n",
    "breezy.bzr.branch",
    "BzrBranchFormat8",
)
branch_format_registry.register_lazy(
    b"Bazaar-NG Branch Reference Format 1\n",
    "breezy.bzr.branch",
    "BranchReferenceFormat",
)

branch_format_registry.set_default_key(b"Bazaar Branch Format 7 (needs bzr 1.6)\n")


# formats which have no format string are not discoverable or independently
# creatable on disk, so are not registered in format_registry.  They're
# all in breezy.bzr.knitreponow.  When an instance of one of these is
# needed, it's constructed directly by the ControlDir.  Non-native formats where
# the repository is not separately opened are similar.

repository_format_registry.register_lazy(
    b"Bazaar-NG Knit Repository Format 1",
    "breezy.bzr.knitrepo",
    "RepositoryFormatKnit1",
)

repository_format_registry.register_lazy(
    b"Bazaar Knit Repository Format 3 (bzr 0.15)\n",
    "breezy.bzr.knitrepo",
    "RepositoryFormatKnit3",
)

repository_format_registry.register_lazy(
    b"Bazaar Knit Repository Format 4 (bzr 1.0)\n",
    "breezy.bzr.knitrepo",
    "RepositoryFormatKnit4",
)

# Pack-based formats. There is one format for pre-subtrees, and one for
# post-subtrees to allow ease of testing.
# NOTE: These are experimental in 0.92. Stable in 1.0 and above
repository_format_registry.register_lazy(
    b"Bazaar pack repository format 1 (needs bzr 0.92)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack1",
)
repository_format_registry.register_lazy(
    b"Bazaar pack repository format 1 with subtree support (needs bzr 0.92)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack3",
)
repository_format_registry.register_lazy(
    b"Bazaar pack repository format 1 with rich root (needs bzr 1.0)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack4",
)
repository_format_registry.register_lazy(
    b"Bazaar RepositoryFormatKnitPack5 (bzr 1.6)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack5",
)
repository_format_registry.register_lazy(
    b"Bazaar RepositoryFormatKnitPack5RichRoot (bzr 1.6.1)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack5RichRoot",
)
repository_format_registry.register_lazy(
    b"Bazaar RepositoryFormatKnitPack5RichRoot (bzr 1.6)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack5RichRootBroken",
)
repository_format_registry.register_lazy(
    b"Bazaar RepositoryFormatKnitPack6 (bzr 1.9)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack6",
)
repository_format_registry.register_lazy(
    b"Bazaar RepositoryFormatKnitPack6RichRoot (bzr 1.9)\n",
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatKnitPack6RichRoot",
)
repository_format_registry.register_lazy(
    b"Bazaar repository format 2a (needs bzr 1.16 or later)\n",
    "breezy.bzr.groupcompress_repo",
    "RepositoryFormat2a",
)

# Development formats.
# Check their docstrings to see if/when they are obsolete.
repository_format_registry.register_lazy(
    (
        b"Bazaar development format 2 with subtree support "
        b"(needs bzr.dev from before 1.8)\n"
    ),
    "breezy.bzr.knitpack_repo",
    "RepositoryFormatPackDevelopment2Subtree",
)
repository_format_registry.register_lazy(
    b"Bazaar development format 8\n",
    "breezy.bzr.groupcompress_repo",
    "RepositoryFormat2aSubtree",
)


workingtree_format_registry.register_lazy(
    b"Bazaar Working Tree Format 4 (bzr 0.15)\n",
    "breezy.bzr.workingtree_4",
    "WorkingTreeFormat4",
)
workingtree_format_registry.register_lazy(
    b"Bazaar Working Tree Format 5 (bzr 1.11)\n",
    "breezy.bzr.workingtree_4",
    "WorkingTreeFormat5",
)
workingtree_format_registry.register_lazy(
    b"Bazaar Working Tree Format 6 (bzr 1.14)\n",
    "breezy.bzr.workingtree_4",
    "WorkingTreeFormat6",
)
workingtree_format_registry.register_lazy(
    b"Bazaar-NG Working Tree format 3", "breezy.bzr.workingtree_3", "WorkingTreeFormat3"
)
workingtree_format_registry.set_default_key(
    b"Bazaar Working Tree Format 6 (bzr 1.14)\n"
)


register_metadir(
    controldir.format_registry,
    "knit",
    "breezy.bzr.knitrepo.RepositoryFormatKnit1",
    "Format using knits.  Recommended for interoperation with bzr <= 0.14.",
    branch_format="breezy.bzr.fullhistory.BzrBranchFormat5",
    tree_format="breezy.bzr.workingtree_3.WorkingTreeFormat3",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "dirstate",
    "breezy.bzr.knitrepo.RepositoryFormatKnit1",
    help="Format using dirstate for working trees. "
    "Compatible with bzr 0.8 and "
    "above when accessed over the network. Introduced in bzr 0.15.",
    branch_format="breezy.bzr.fullhistory.BzrBranchFormat5",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "dirstate-tags",
    "breezy.bzr.knitrepo.RepositoryFormatKnit1",
    help="Variant of dirstate with support for tags. Introduced in bzr 0.15.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "rich-root",
    "breezy.bzr.knitrepo.RepositoryFormatKnit4",
    help="Variant of dirstate with better handling of tree roots. "
    "Introduced in bzr 1.0.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "dirstate-with-subtree",
    "breezy.bzr.knitrepo.RepositoryFormatKnit3",
    help="Variant of dirstate with support for nested trees. Introduced in 0.15.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    experimental=True,
    deprecated=True,
    hidden=True,
)
register_metadir(
    controldir.format_registry,
    "pack-0.92",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack1",
    help="Pack-based format used in 1.x series. Introduced in 0.92. "
    "Interoperates with bzr repositories before 0.92 but cannot be "
    "read by bzr < 0.92.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    deprecated=True,
    hidden=True,
)
register_metadir(
    controldir.format_registry,
    "pack-0.92-subtree",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack3",
    help="Pack-based format used in 1.x series, with subtree support. "
    "Introduced in 0.92. Interoperates with "
    "bzr repositories before 0.92 but cannot be read by bzr < 0.92.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
    experimental=True,
)
register_metadir(
    controldir.format_registry,
    "rich-root-pack",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack4",
    help="A variant of pack-0.92 that supports rich-root data "
    "(needed for bzr-svn and bzr-git). Introduced in 1.0.",
    branch_format="breezy.bzr.branch.BzrBranchFormat6",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.6",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack5",
    help="A format that allows a branch to indicate that there is another "
    "(stacked) repository that should be used to access data that is "
    "not present locally.",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.6.1-rich-root",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack5RichRoot",
    help="A variant of 1.6 that supports rich-root data "
    "(needed for bzr-svn and bzr-git).",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.9",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6",
    help="A repository format using B+tree indexes. These indexes "
    "are smaller in size, have smarter caching and provide faster "
    "performance for most operations.",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.9-rich-root",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6RichRoot",
    help="A variant of 1.9 that supports rich-root data "
    "(needed for bzr-svn and bzr-git).",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat4",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.14",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6",
    help="A working-tree format that supports content filtering.",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat5",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "1.14-rich-root",
    "breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6RichRoot",
    help="A variant of 1.14 that supports rich-root data "
    "(needed for bzr-svn and bzr-git).",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat5",
    hidden=True,
    deprecated=True,
)
# The following un-numbered 'development' formats should always just be aliases.
register_metadir(
    controldir.format_registry,
    "development-subtree",
    "breezy.bzr.groupcompress_repo.RepositoryFormat2aSubtree",
    help="Current development format, subtree variant. Can convert data to and "
    "from pack-0.92-subtree (and anything compatible with "
    "pack-0.92-subtree) format repositories. Repositories and branches in "
    "this format can only be read by bzr.dev. Please read "
    "https://www.breezy-vcs.org/developers/development-repo.html "
    "before use.",
    branch_format="breezy.bzr.branch.BzrBranchFormat8",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat6",
    experimental=True,
    deprecated=True,
    hidden=True,
)
register_metadir(
    controldir.format_registry,
    "development5-subtree",
    "breezy.bzr.knitpack_repo.RepositoryFormatPackDevelopment2Subtree",
    help="Development format, subtree variant. Can convert data to and "
    "from pack-0.92-subtree (and anything compatible with "
    "pack-0.92-subtree) format repositories. Repositories and branches in "
    "this format can only be read by bzr.dev. Please read "
    "https://www.breezy-vcs.org/developers/development-repo.html "
    "before use.",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat6",
    experimental=True,
    deprecated=True,
    hidden=True,
)

register_metadir(
    controldir.format_registry,
    "development-colo",
    "breezy.bzr.groupcompress_repo.RepositoryFormat2a",
    help="The 2a format with experimental support for colocated branches.\n",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat6",
    experimental=True,
    bzrdir_format="breezy.bzr.bzrdir.BzrDirMetaFormat1Colo",
    hidden=True,
)


# And the development formats above will have aliased one of the following:

# Finally, the current format.
register_metadir(
    controldir.format_registry,
    "2a",
    "breezy.bzr.groupcompress_repo.RepositoryFormat2a",
    help="Format for the bzr 2.0 series.\n",
    branch_format="breezy.bzr.branch.BzrBranchFormat7",
    tree_format="breezy.bzr.workingtree_4.WorkingTreeFormat6",
    experimental=False,
    deprecated=False,
)

# The following format should be an alias for the rich root equivalent
# of the default format

controldir.format_registry.register_alias("default-rich-root", "2a", hidden=True)

# The following format should is just an alias for the default bzr format.
controldir.format_registry.register_alias("bzr", "2a")

# The current format that is made on 'bzr init'.
format_name = config.GlobalStack().get("default_format")
controldir.format_registry.set_default(format_name)



# Copyright (C) 2007, 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import errno
from email.message import Message

from breezy import config, email_message, smtp_connection, tests, ui


def connection_refuser(host):
    raise OSError(errno.ECONNREFUSED, "Connection Refused")


class StubSMTPFactory:
    """A fake SMTP connection to test the connection setup."""

    def __init__(self, fail_on=None, smtp_features=None):
        self._fail_on = fail_on or []
        self._calls = []
        self._smtp_features = smtp_features or []
        self._ehlo_called = False

    def __call__(self, host="localhost"):
        self._calls.append(("connect", host))
        # The factory pretends to be a connection
        return self

    def connect(self, server):
        raise NotImplementedError

    def helo(self):
        self._calls.append(("helo",))
        if "helo" in self._fail_on:
            return 500, "helo failure"
        else:
            return 200, "helo success"

    def ehlo(self):
        self._calls.append(("ehlo",))
        if "ehlo" in self._fail_on:
            return 500, "ehlo failure"
        else:
            self._ehlo_called = True
            return 200, "ehlo success"

    def has_extn(self, extension):
        self._calls.append(("has_extn", extension))
        return self._ehlo_called and extension in self._smtp_features

    def starttls(self):
        self._calls.append(("starttls",))
        if "starttls" in self._fail_on:
            return 500, "starttls failure"
        else:
            self._ehlo_called = True
            return 200, "starttls success"


class WideOpenSMTPFactory(StubSMTPFactory):
    """A fake smtp server that implements login by accepting anybody."""

    def login(self, user, password):
        self._calls.append(("login", user, password))


class TestSMTPConnection(tests.TestCaseInTempDir):
    def get_connection(self, text, smtp_factory=None):
        my_config = config.MemoryStack(text)
        return smtp_connection.SMTPConnection(my_config, _smtp_factory=smtp_factory)

    def test_defaults(self):
        conn = self.get_connection(b"")
        self.assertEqual("localhost", conn._smtp_server)
        self.assertEqual(None, conn._smtp_username)
        self.assertEqual(None, conn._smtp_password)

    def test_smtp_server(self):
        conn = self.get_connection(b"smtp_server=host:10")
        self.assertEqual("host:10", conn._smtp_server)

    def test_missing_server(self):
        conn = self.get_connection(b"", smtp_factory=connection_refuser)
        self.assertRaises(smtp_connection.DefaultSMTPConnectionRefused, conn._connect)
        conn = self.get_connection(
            b"smtp_server=smtp.example.com", smtp_factory=connection_refuser
        )
        self.assertRaises(smtp_connection.SMTPConnectionRefused, conn._connect)

    def test_smtp_username(self):
        conn = self.get_connection(b"")
        self.assertIs(None, conn._smtp_username)

        conn = self.get_connection(b"smtp_username=joebody")
        self.assertEqual("joebody", conn._smtp_username)

    def test_smtp_password_from_config(self):
        conn = self.get_connection(b"")
        self.assertIs(None, conn._smtp_password)

        conn = self.get_connection(b"smtp_password=mypass")
        self.assertEqual("mypass", conn._smtp_password)

    def test_smtp_password_from_user(self):
        user = "joe"
        password = "hispass"
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b"[DEFAULT]\nsmtp_username=%s\n" % user.encode("ascii"),
            smtp_factory=factory,
        )
        self.assertIs(None, conn._smtp_password)

        ui.ui_factory = ui.CannedInputUIFactory([password])
        conn._connect()
        self.assertEqual(password, conn._smtp_password)

    def test_smtp_password_from_auth_config(self):
        user = "joe"
        password = "hispass"
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b"[DEFAULT]\nsmtp_username=%s\n" % user.encode("ascii"),
            smtp_factory=factory,
        )
        self.assertEqual(user, conn._smtp_username)
        self.assertIs(None, conn._smtp_password)
        # Create a config file with the right password
        conf = config.AuthenticationConfig()
        conf._get_config().update(
            {"smtptest": {"scheme": "smtp", "user": user, "password": password}}
        )
        conf._save()

        conn._connect()
        self.assertEqual(password, conn._smtp_password)

    def test_authenticate_with_byte_strings(self):
        user = b"joe"
        unicode_pass = "h\xecspass"
        utf8_pass = unicode_pass.encode("utf-8")
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b"[DEFAULT]\nsmtp_username=%s\nsmtp_password=%s\n" % (user, utf8_pass),
            smtp_factory=factory,
        )
        self.assertEqual(unicode_pass, conn._smtp_password)
        conn._connect()
        self.assertEqual(
            [
                ("connect", "localhost"),
                ("ehlo",),
                ("has_extn", "starttls"),
                ("login", user, utf8_pass),
            ],
            factory._calls,
        )
        smtp_username, smtp_password = factory._calls[-1][1:]
        self.assertIsInstance(smtp_username, bytes)
        self.assertIsInstance(smtp_password, bytes)

    def test_create_connection(self):
        factory = StubSMTPFactory()
        conn = self.get_connection(b"", smtp_factory=factory)
        conn._create_connection()
        self.assertEqual(
            [("connect", "localhost"), ("ehlo",), ("has_extn", "starttls")],
            factory._calls,
        )

    def test_create_connection_ehlo_fails(self):
        # Check that we call HELO if EHLO failed.
        factory = StubSMTPFactory(fail_on=["ehlo"])
        conn = self.get_connection(b"", smtp_factory=factory)
        conn._create_connection()
        self.assertEqual(
            [("connect", "localhost"), ("ehlo",), ("helo",), ("has_extn", "starttls")],
            factory._calls,
        )

    def test_create_connection_ehlo_helo_fails(self):
        # Check that we raise an exception if both EHLO and HELO fail.
        factory = StubSMTPFactory(fail_on=["ehlo", "helo"])
        conn = self.get_connection(b"", smtp_factory=factory)
        self.assertRaises(smtp_connection.SMTPError, conn._create_connection)
        self.assertEqual(
            [("connect", "localhost"), ("ehlo",), ("helo",)], factory._calls
        )

    def test_create_connection_starttls(self):
        # Check that STARTTLS plus a second EHLO are called if the
        # server says it supports the feature.
        factory = StubSMTPFactory(smtp_features=["starttls"])
        conn = self.get_connection(b"", smtp_factory=factory)
        conn._create_connection()
        self.assertEqual(
            [
                ("connect", "localhost"),
                ("ehlo",),
                ("has_extn", "starttls"),
                ("starttls",),
                ("ehlo",),
            ],
            factory._calls,
        )

    def test_create_connection_starttls_fails(self):
        # Check that we raise an exception if the server claims to
        # support STARTTLS, but then fails when we try to activate it.
        factory = StubSMTPFactory(fail_on=["starttls"], smtp_features=["starttls"])
        conn = self.get_connection(b"", smtp_factory=factory)
        self.assertRaises(smtp_connection.SMTPError, conn._create_connection)
        self.assertEqual(
            [
                ("connect", "localhost"),
                ("ehlo",),
                ("has_extn", "starttls"),
                ("starttls",),
            ],
            factory._calls,
        )

    def test_get_message_addresses(self):
        msg = Message()

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual("", from_)
        self.assertEqual([], to)

        msg["From"] = '"J. Random Developer" <jrandom@example.com>'
        msg["To"] = "John Doe <john@doe.com>, Jane Doe <jane@doe.com>"
        msg["CC"] = "Pepe P\xe9rez <pperez@ejemplo.com>"
        msg["Bcc"] = "user@localhost"

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual("jrandom@example.com", from_)
        self.assertEqual(
            sorted(
                ["john@doe.com", "jane@doe.com", "pperez@ejemplo.com", "user@localhost"]
            ),
            sorted(to),
        )

        # now with breezy's EmailMessage
        msg = email_message.EmailMessage(
            '"J. Random Developer" <jrandom@example.com>',
            [
                "John Doe <john@doe.com>",
                "Jane Doe <jane@doe.com>",
                "Pepe P\xe9rez <pperez@ejemplo.com>",
                "user@localhost",
            ],
            "subject",
        )

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual("jrandom@example.com", from_)
        self.assertEqual(
            sorted(
                ["john@doe.com", "jane@doe.com", "pperez@ejemplo.com", "user@localhost"]
            ),
            sorted(to),
        )

    def test_destination_address_required(self):
        msg = Message()
        msg["From"] = '"J. Random Developer" <jrandom@example.com>'
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")).send_email,
            msg,
        )

        msg = email_message.EmailMessage("from@from.com", "", "subject")
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")).send_email,
            msg,
        )

        msg = email_message.EmailMessage("from@from.com", [], "subject")
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")).send_email,
            msg,
        )



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#


"""Tests of the 'brz clean-tree' command."""

from breezy import ignores
from breezy.tests import TestCaseWithTransport
from breezy.tests.script import run_script


class TestBzrTools(TestCaseWithTransport):
    @staticmethod
    def touch(filename):
        with open(filename, "wb") as my_file:
            my_file.write(b"")

    def test_clean_tree(self):
        self.run_bzr("init")
        self.run_bzr("ignore *~")
        self.run_bzr("ignore *.pyc")
        self.touch("name")
        self.touch("name~")
        self.assertPathExists("name~")
        self.touch("name.pyc")
        self.run_bzr("clean-tree --force")
        self.assertPathExists("name~")
        self.assertPathDoesNotExist("name")
        self.touch("name")
        self.run_bzr("clean-tree --detritus --force")
        self.assertPathExists("name")
        self.assertPathDoesNotExist("name~")
        self.assertPathExists("name.pyc")
        self.run_bzr("clean-tree --ignored --force")
        self.assertPathExists("name")
        self.assertPathDoesNotExist("name.pyc")
        self.run_bzr("clean-tree --unknown --force")
        self.assertPathDoesNotExist("name")
        self.touch("name")
        self.touch("name~")
        self.touch("name.pyc")
        self.run_bzr("clean-tree --unknown --ignored --force")
        self.assertPathDoesNotExist("name")
        self.assertPathDoesNotExist("name~")
        self.assertPathDoesNotExist("name.pyc")

    def test_clean_tree_nested_bzrdir(self):
        # clean-tree should not blindly delete nested bzrdirs (branches)
        # bug https://bugs.launchpad.net/bzr/+bug/572098
        # so it will play well with scmproj/bzr-externals plugins.
        wt1 = self.make_branch_and_tree(".")
        self.make_branch_and_tree("foo")
        self.make_branch_and_tree("bar")
        ignores.tree_ignores_add_patterns(wt1, ["./foo"])
        self.run_bzr(["clean-tree", "--unknown", "--force"])
        self.assertPathExists("foo")
        self.assertPathExists("bar")
        self.run_bzr(["clean-tree", "--ignored", "--force"])
        self.assertPathExists("foo")
        self.assertPathExists("bar")

    def test_clean_tree_directory(self):
        """Test --directory option."""
        tree = self.make_branch_and_tree("a")
        self.build_tree(["a/added", "a/unknown", "a/ignored"])
        tree.add("added")
        self.run_bzr("clean-tree -d a --unknown --ignored --force")
        self.assertPathDoesNotExist("a/unknown")
        self.assertPathDoesNotExist("a/ignored")
        self.assertPathExists("a/added")

    def test_clean_tree_interactive(self):
        self.make_branch_and_tree(".")
        self.touch("bar")
        self.touch("foo")
        run_script(
            self,
            """
        $ brz clean-tree
        bar
        foo
        2>Are you sure you wish to delete these? ([y]es, [n]o): no
        <n
        Canceled
        """,
        )
        self.assertPathExists("bar")
        self.assertPathExists("foo")
        run_script(
            self,
            """
        $ brz clean-tree
        bar
        foo
        2>Are you sure you wish to delete these? ([y]es, [n]o): yes
        <y
        2>deleting paths:
        2>  bar
        2>  foo
        """,
        )
        self.assertPathDoesNotExist("bar")
        self.assertPathDoesNotExist("foo")



# Copyright (C) 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import tests, workingtree
from breezy.bzr.knitpack_repo import RepositoryFormatKnitPack4
from breezy.bzr.knitrepo import RepositoryFormatKnit4


class TestSplit(tests.TestCaseWithTransport):
    def test_split(self):
        self.build_tree(["a/", "a/b/", "a/b/c"])
        wt = self.make_branch_and_tree("a", format="rich-root-pack")
        wt.add(["b", "b/c"])
        wt.commit("rev1")
        self.run_bzr("split a/b")
        self.run_bzr_error((".* is not versioned",), "split q", working_dir="a")

    def test_split_repo_failure(self):
        repo = self.make_repository("branch", shared=True, format="knit")
        a_branch = repo.controldir.create_branch()
        self.build_tree(["a/", "a/b/", "a/b/c/", "a/b/c/d"])
        wt = a_branch.create_checkout("a", lightweight=True)
        wt.add(["b", "b/c", "b/c/d"], ids=[b"b-id", b"c-id", b"d-id"])
        wt.commit("added files")
        self.run_bzr_error(("must upgrade your branch at .*a",), "split a/b")

    def test_split_tree_failure(self):
        tree = self.make_branch_and_tree("tree", format="pack-0.92")
        self.build_tree(["tree/subtree/"])
        tree.add("subtree")
        tree.commit("added subtree")
        self.run_bzr_error(
            ("must upgrade your branch at .*tree", "rich roots"), "split tree/subtree"
        )

    def split_formats(self, format, repo_format):
        tree = self.make_branch_and_tree("rich-root", format=format)
        self.build_tree(["rich-root/a/"])
        tree.add("a")
        self.run_bzr(["split", "rich-root/a"])
        subtree = workingtree.WorkingTree.open("rich-root/a")
        self.assertIsInstance(subtree.branch.repository._format, repo_format)

    def test_split_rich_root(self):
        self.split_formats("rich-root", RepositoryFormatKnit4)

    def test_split_rich_root_pack(self):
        self.split_formats("rich-root-pack", RepositoryFormatKnitPack4)



# Copyright (C) 2006, 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Adapter for running test cases against multiple encodings."""

# prefix for micro (1/1000000)
_mu = "\xb5"

# greek letter omega, not to be confused with
# the Ohm sign, u'\u2126'. Though they are probably identical
# cp437 can handle the first, but not the second
_omega = "\u03a9"

# smallest error possible, epsilon
# cp437 handles u03b5, but not u2208 the 'element of' operator
_epsilon = "\u03b5"

# Swedish?
_erik = "Erik B\xe5gfors"

# Swedish 'räksmörgås' means shrimp sandwich
_shrimp_sandwich = "r\xe4ksm\xf6rg\xe5s"

# Arabic, probably only Unicode encodings can handle this one
_juju = "\u062c\u0648\u062c\u0648"

# iso-8859-1 alternative for juju
_juju_alt = "j\xfbj\xfa"

# Russian, 'Alexander' in russian
_alexander = "\u0410\u043b\u0435\u043a\u0441\u0430\u043d\u0434\u0440"
# The word 'test' in Russian
_russian_test = "\u0422\u0435\u0441\u0442"

# Kanji
# It is a kanji sequence for nihonjin, or Japanese in English.
#
# '\u4eba' being person, 'u\65e5' sun and '\u672c' origin. Ie,
# sun-origin-person, 'native from the land where the sun rises'. Note, I'm
# not a fluent speaker, so this is just my crude breakdown.
#
# Wouter van Heyst
_nihonjin = "\u65e5\u672c\u4eba"

# Czech
# It's what is usually used for showing how fonts look, because it contains
# most accented characters, ie. in places where Englishman use 'Quick brown fox
# jumped over a lazy dog'. The literal translation of the Czech version would
# be something like 'Yellow horse groaned devilish codes'. Actually originally
# the last word used to be 'ódy' (odes). The 'k' was added as a pun when using
# the sentece to check whether one has properly set encoding.
_yellow_horse = (
    "\u017dlu\u0165ou\u010dk\xfd k\u016f\u0148 \xfap\u011bl \u010f\xe1belsk\xe9 k\xf3dy"
)
_yellow = "\u017dlu\u0165ou\u010dk\xfd"
_someone = "Some\u016f\u0148\u011b"
_something = "\u0165ou\u010dk\xfd"

# Hebrew
# Shalom -> 'hello' or 'peace', used as a common greeting
_shalom = "\u05e9\u05dc\u05d5\u05dd"


encoding_scenarios = [
    # Permutation 1 of utf-8
    (
        "utf-8,1",
        {
            "info": {
                "committer": _erik,
                "message": _yellow_horse,
                "filename": _shrimp_sandwich,
                "directory": _nihonjin,
            },
            "encoding": "utf-8",
        },
    ),
    # Permutation 2 of utf-8
    (
        "utf-8,2",
        {
            "info": {
                "committer": _alexander,
                "message": "Testing " + _mu,
                "filename": _shalom,
                "directory": _juju,
            },
            "encoding": "utf-8",
        },
    ),
    (
        "iso-8859-1",
        {
            "info": {
                "committer": _erik,
                "message": "Testing " + _mu,
                "filename": _juju_alt,
                "directory": _shrimp_sandwich,
            },
            "encoding": "iso-8859-1",
        },
    ),
    (
        "iso-8859-2",
        {
            "info": {
                "committer": _someone,
                "message": _yellow_horse,
                "filename": _yellow,
                "directory": _something,
            },
            "encoding": "iso-8859-2",
        },
    ),
    (
        "cp1251",
        {
            "info": {
                "committer": _alexander,
                "message": "Testing " + _mu,
                "filename": _russian_test,
                "directory": _russian_test + "dir",
            },
            "encoding": "cp1251",
        },
    ),
    # The iso-8859-1 tests run on a default windows cp437 installation
    # and it takes a long time to run an extra permutation of the tests
    # But just in case we want to add this back in:
    #        ('cp437', {'committer':_erik
    #                  , 'message':u'Testing ' + _mu
    #                  , 'filename':'file_' + _omega
    #                  , 'directory':_epsilon + '_dir',
    #            'encoding': 'cp437'}),
]



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ... import graph as _mod_graph
from ... import tests
from ...revision import NULL_REVISION
from ...tests.test_graph import TestGraphBase
from .. import vf_search

# Ancestry 1:
#
#  NULL_REVISION
#       |
#     rev1
#      /\
#  rev2a rev2b
#     |    |
#   rev3  /
#     |  /
#   rev4
ancestry_1 = {
    b"rev1": [NULL_REVISION],
    b"rev2a": [b"rev1"],
    b"rev2b": [b"rev1"],
    b"rev3": [b"rev2a"],
    b"rev4": [b"rev3", b"rev2b"],
}

# Ancestry 2:
#
#  NULL_REVISION
#    /    \
# rev1a  rev1b
#   |
# rev2a
#   |
# rev3a
#   |
# rev4a
ancestry_2 = {
    b"rev1a": [NULL_REVISION],
    b"rev2a": [b"rev1a"],
    b"rev1b": [NULL_REVISION],
    b"rev3a": [b"rev2a"],
    b"rev4a": [b"rev3a"],
}


# Extended history shortcut
#  NULL_REVISION
#       |
#       a
#       |\
#       b |
#       | |
#       c |
#       | |
#       d |
#       |\|
#       e f
extended_history_shortcut = {
    b"a": [NULL_REVISION],
    b"b": [b"a"],
    b"c": [b"b"],
    b"d": [b"c"],
    b"e": [b"d"],
    b"f": [b"a", b"d"],
}


class TestSearchResultRefine(tests.TestCase):
    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def test_refine(self):
        # Used when pulling from a stacked repository, so test some revisions
        # being satisfied from the stacking branch.
        self.make_graph(
            {
                b"tip": [b"mid"],
                b"mid": [b"base"],
                b"tag": [b"base"],
                b"base": [NULL_REVISION],
                NULL_REVISION: [],
            }
        )
        result = vf_search.SearchResult(
            {b"tip", b"tag"}, {NULL_REVISION}, 4, {b"tip", b"mid", b"tag", b"base"}
        )
        result = result.refine({b"tip"}, {b"mid"})
        recipe = result.get_recipe()
        # We should be starting from tag (original head) and mid (seen ref)
        self.assertEqual({b"mid", b"tag"}, recipe[1])
        # We should be stopping at NULL (original stop) and tip (seen head)
        self.assertEqual({NULL_REVISION, b"tip"}, recipe[2])
        self.assertEqual(3, recipe[3])
        result = result.refine({b"mid", b"tag", b"base"}, {NULL_REVISION})
        recipe = result.get_recipe()
        # We should be starting from nothing (NULL was known as a cut point)
        self.assertEqual(set(), recipe[1])
        # We should be stopping at NULL (original stop) and tip (seen head) and
        # tag (seen head) and mid(seen mid-point head). We could come back and
        # define this as not including mid, for minimal results, but it is
        # still 'correct' to include mid, and simpler/easier.
        self.assertEqual({NULL_REVISION, b"tip", b"tag", b"mid"}, recipe[2])
        self.assertEqual(0, recipe[3])
        self.assertTrue(result.is_empty())


class TestSearchResultFromParentMap(TestGraphBase):
    def assertSearchResult(
        self, start_keys, stop_keys, key_count, parent_map, missing_keys=()
    ):
        (start, stop, count) = vf_search.search_result_from_parent_map(
            parent_map, missing_keys
        )
        self.assertEqual(
            (sorted(start_keys), sorted(stop_keys), key_count),
            (sorted(start), sorted(stop), count),
        )

    def test_no_parents(self):
        self.assertSearchResult([], [], 0, {})
        self.assertSearchResult([], [], 0, None)

    def test_ancestry_1(self):
        self.assertSearchResult([b"rev4"], [NULL_REVISION], len(ancestry_1), ancestry_1)

    def test_ancestry_2(self):
        self.assertSearchResult(
            [b"rev1b", b"rev4a"], [NULL_REVISION], len(ancestry_2), ancestry_2
        )
        self.assertSearchResult(
            [b"rev1b", b"rev4a"],
            [],
            len(ancestry_2) + 1,
            ancestry_2,
            missing_keys=[NULL_REVISION],
        )

    def test_partial_search(self):
        parent_map = {k: extended_history_shortcut[k] for k in [b"e", b"f"]}
        self.assertSearchResult([b"e", b"f"], [b"d", b"a"], 2, parent_map)
        parent_map.update((k, extended_history_shortcut[k]) for k in [b"d", b"a"])
        self.assertSearchResult([b"e", b"f"], [b"c", NULL_REVISION], 4, parent_map)
        parent_map[b"c"] = extended_history_shortcut[b"c"]
        self.assertSearchResult(
            [b"e", b"f"], [b"b"], 6, parent_map, missing_keys=[NULL_REVISION]
        )
        parent_map[b"b"] = extended_history_shortcut[b"b"]
        self.assertSearchResult(
            [b"e", b"f"], [], 7, parent_map, missing_keys=[NULL_REVISION]
        )


class TestLimitedSearchResultFromParentMap(TestGraphBase):
    def assertSearchResult(
        self,
        start_keys,
        stop_keys,
        key_count,
        parent_map,
        missing_keys,
        tip_keys,
        depth,
    ):
        (start, stop, count) = vf_search.limited_search_result_from_parent_map(
            parent_map, missing_keys, tip_keys, depth
        )
        self.assertEqual(
            (sorted(start_keys), sorted(stop_keys), key_count),
            (sorted(start), sorted(stop), count),
        )

    def test_empty_ancestry(self):
        self.assertSearchResult([], [], 0, {}, (), [b"tip-rev-id"], 10)

    def test_ancestry_1(self):
        self.assertSearchResult([b"rev4"], [b"rev1"], 4, ancestry_1, (), [b"rev1"], 10)
        self.assertSearchResult(
            [b"rev2a", b"rev2b"], [b"rev1"], 2, ancestry_1, (), [b"rev1"], 1
        )

    def test_multiple_heads(self):
        self.assertSearchResult(
            [b"e", b"f"], [b"a"], 5, extended_history_shortcut, (), [b"a"], 10
        )
        # Note that even though we only take 1 step back, we find 'f', which
        # means the described search will still find d and c.
        self.assertSearchResult(
            [b"f"], [b"a"], 4, extended_history_shortcut, (), [b"a"], 1
        )
        self.assertSearchResult(
            [b"f"], [b"a"], 4, extended_history_shortcut, (), [b"a"], 2
        )


class TestPendingAncestryResultRefine(tests.TestCase):
    def make_graph(self, ancestors):
        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))

    def test_refine(self):
        # Used when pulling from a stacked repository, so test some revisions
        # being satisfied from the stacking branch.
        self.make_graph(
            {
                b"tip": [b"mid"],
                b"mid": [b"base"],
                b"tag": [b"base"],
                b"base": [NULL_REVISION],
                NULL_REVISION: [],
            }
        )
        result = vf_search.PendingAncestryResult([b"tip", b"tag"], None)
        result = result.refine({b"tip"}, {b"mid"})
        self.assertEqual({b"mid", b"tag"}, result.heads)
        result = result.refine({b"mid", b"tag", b"base"}, {NULL_REVISION})
        self.assertEqual({NULL_REVISION}, result.heads)
        self.assertTrue(result.is_empty())


class TestPendingAncestryResultGetKeys(tests.TestCaseWithMemoryTransport):
    """Tests for breezy.graph.PendingAncestryResult."""

    def test_get_keys(self):
        builder = self.make_branch_builder("b")
        builder.start_series()
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", ""))], revision_id=b"rev-1"
        )
        builder.build_snapshot([b"rev-1"], [], revision_id=b"rev-2")
        builder.finish_series()
        repo = builder.get_branch().repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        result = vf_search.PendingAncestryResult([b"rev-2"], repo)
        self.assertEqual({b"rev-1", b"rev-2"}, set(result.get_keys()))

    def test_get_keys_excludes_ghosts(self):
        builder = self.make_branch_builder("b")
        builder.start_series()
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", ""))], revision_id=b"rev-1"
        )
        builder.build_snapshot([b"rev-1", b"ghost"], [], revision_id=b"rev-2")
        builder.finish_series()
        repo = builder.get_branch().repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        result = vf_search.PendingAncestryResult([b"rev-2"], repo)
        self.assertEqual(sorted([b"rev-1", b"rev-2"]), sorted(result.get_keys()))

    def test_get_keys_excludes_null(self):
        # Make a 'graph' with an iter_ancestry that returns NULL_REVISION
        # somewhere other than the last element, which can happen in real
        # ancestries.
        class StubGraph:
            def iter_ancestry(self, keys):
                return [(NULL_REVISION, ()), (b"foo", (NULL_REVISION,))]

        result = vf_search.PendingAncestryResult([b"rev-3"], None)
        result_keys = result._get_keys(StubGraph())
        # Only the non-null keys from the ancestry appear.
        self.assertEqual({b"foo"}, set(result_keys))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of canonical paths of trees."""

from breezy.tests import features
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestCanonicalPaths(TestCaseWithWorkingTree):
    def _make_canonical_test_tree(self, commit=True):
        # make a tree used by all the 'canonical' tests below.
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/dir/", "tree/dir/file"])
        work_tree.add(["dir", "dir/file"])
        if commit:
            work_tree.commit("commit 1")
        # XXX: this isn't actually guaranteed to return the class we want to
        # test -- mbp 2010-02-12
        return work_tree

    def test_canonical_path(self):
        work_tree = self._make_canonical_test_tree()
        if features.CaseInsensitiveFilesystemFeature.available():
            self.assertEqual("dir/file", work_tree.get_canonical_path("Dir/File"))
        elif features.CaseInsCasePresFilenameFeature.available():
            self.assertEqual("dir/file", work_tree.get_canonical_path("Dir/File"))
        else:
            self.assertEqual("Dir/File", work_tree.get_canonical_path("Dir/File"))

    def test_canonical_path_before_commit(self):
        work_tree = self._make_canonical_test_tree(False)
        if features.CaseInsensitiveFilesystemFeature.available():
            self.assertEqual("dir/file", work_tree.get_canonical_path("Dir/File"))
        elif features.CaseInsCasePresFilenameFeature.available():
            self.assertEqual("dir/file", work_tree.get_canonical_path("Dir/File"))
        else:
            self.assertEqual("Dir/File", work_tree.get_canonical_path("Dir/File"))

    def test_canonical_path_dir(self):
        # check it works when asked for just the directory portion.
        work_tree = self._make_canonical_test_tree()
        if features.CaseInsensitiveFilesystemFeature.available():
            self.assertEqual("dir", work_tree.get_canonical_path("Dir"))
        elif features.CaseInsCasePresFilenameFeature.available():
            self.assertEqual("dir", work_tree.get_canonical_path("Dir"))
        else:
            self.assertEqual("Dir", work_tree.get_canonical_path("Dir"))

    def test_canonical_path_root(self):
        work_tree = self._make_canonical_test_tree()
        self.assertEqual("", work_tree.get_canonical_path(""))
        self.assertEqual("", work_tree.get_canonical_path("/"))

    def test_canonical_path_invalid_all(self):
        work_tree = self._make_canonical_test_tree()
        self.assertEqual("foo/bar", work_tree.get_canonical_path("foo/bar"))

    def test_canonical_invalid_child(self):
        work_tree = self._make_canonical_test_tree()
        if features.CaseInsensitiveFilesystemFeature.available():
            self.assertEqual("dir/None", work_tree.get_canonical_path("Dir/None"))
        elif features.CaseInsCasePresFilenameFeature.available():
            self.assertEqual("dir/None", work_tree.get_canonical_path("Dir/None"))
        else:
            self.assertEqual("Dir/None", work_tree.get_canonical_path("Dir/None"))

    def test_canonical_tree_name_mismatch(self):
        # see <https://bugs.launchpad.net/bzr/+bug/368931>
        # some of the trees we want to use can only exist on a disk, not in
        # memory - therefore we can only test this if the filesystem is
        # case-sensitive.
        self.requireFeature(features.case_sensitive_filesystem_feature)
        work_tree = self.make_branch_and_tree(".")
        self.build_tree(["test/", "test/file", "Test"])
        work_tree.add(["test/", "test/file", "Test"])

        self.assertEqual(
            ["test", "Test", "test/file", "Test/file"],
            list(
                work_tree.get_canonical_paths(
                    ["test", "Test", "test/file", "Test/file"]
                )
            ),
        )

        test_revid = work_tree.commit("commit")
        test_tree = work_tree.branch.repository.revision_tree(test_revid)
        test_tree.lock_read()
        self.addCleanup(test_tree.unlock)

        self.assertEqual(
            ["", "Test", "test", "test/file"],
            [p for p, e in test_tree.iter_entries_by_dir()],
        )
